from os.path import join, dirname, abspath, exists, splitext
import ayugespidertools
from scrapy.commands.genspider import Command


class AyuCommand(Command):

    @property
    def templates_dir(self):
        return join(
            join(ayugespidertools.__path__[0], 'templates'),
            'spiders'
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@File    :  MultiPlexing.py
@Time    :  2022/7/12 16:53
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  用于存放经常使用的方法
"""
import copy
import json
import re
import os
import cv2
import random
import pymysql
import numpy as np
from typing import Optional, Union
from ayugespidertools.config import logger


__all__ = [
    'ReuseOperation',
]


class ReuseOperation(object):
    """
    用于存放经常复用的一些操作
    """

    @staticmethod
    def judge_file_style(media_file_or_url: str, media_type: Optional[str] = None, strict: bool = False) -> str:
        """
        判断新增图片的格式类型（此方法是为了防止图片地址格式不统一，造成图片格式提取错误）
        Args:
            media_file_or_url: 需要判断格式类型的文件或文件链接
            media_type: media 文件的类型，image，vedio，voice 等
            strict: media 是否为严格模式，严格模式下也需要仔细判断

        Returns:
            1): 文件的格式信息
        """
        if any([str(media_file_or_url).startswith("http"), strict]):
            image_style = ['.svg', '.png', '.jpg', '.jpeg', '.bmp', '.wav', '.mp3', '.ogg', '.flv']
            for image_format in image_style:
                if image_format in media_file_or_url:
                    return image_format[1:]

        # 否则就是文件，则返回文件后缀即可
        return media_file_or_url.split('.')[-1]

    @staticmethod
    def get_file_name_by_url(file_url: str) -> str:
        """
        根据文件链接取出文件的名称
        Args:
            file_url: 文件链接

        Returns:
            1). 文件名称
        """
        pattern = re.compile(r""".*/(.*?)\..*?""")
        file_name_list = pattern.findall(file_url)
        if file_name_list:
            return file_name_list[0]
        return ""

    # TODO: 优化此方法，修改方法名称等
    @staticmethod
    def get_voice_files(path: str) -> list:
        """
        获取 path 文件夹下的所有文件
        Args:
            path: 需要判断的文件夹路径

        Returns:
            file_list: path 文件夹下的文件列表
        """
        # 得到文件夹下的所有文件名称
        files = os.listdir(path)
        file_list = []
        # 遍历该文件夹
        for file in files:
            # 是子文件夹
            if os.path.isdir(path + "\\" + file):
                # get_voice_files(path + "\\" + file)
                continue

            # 是文件
            else:
                file_list.append(file)
        return file_list

    @staticmethod
    def get_bytes_by_file(file_path: str) -> bytes:
        """
        获取媒体文件的 bytes 内容
        Args:
            file_path: 对应文件的路径

        Returns:
            file_bytes: file_path 对应文件的 bytes 内容
        """
        with open(file_path, "rb") as f:
            file_bytes = f.read()
        return file_bytes

    @staticmethod
    def read_image_data(bg, tp):
        if any([type(bg) != str, type(bg) == bytes]):
            bg_buf = np.frombuffer(bg, np.uint8)
            tp_buf = np.frombuffer(tp, np.uint8)

            bg_cv = cv2.imdecode(bg_buf, cv2.IMREAD_ANYCOLOR)
            tp_cv = cv2.imdecode(tp_buf, cv2.IMREAD_ANYCOLOR)

        else:
            # 读取图片，读进来直接是 BGR 格式数据格式在 0~255
            bg_cv = cv2.imread(bg)
            # 0 表示采用黑白的方式读取图片
            tp_cv = cv2.imread(tp, 0)

        return bg_cv, tp_cv

    @staticmethod
    def random_weight(weight_data: list):
        """
        带权重的随机取值，即在带权重的列表数据中根据权重随机取一个值
        Args:
            weight_data: 带权重的列表信息，示例：[{'username': 'xxxx', 'password': '******', 'weight': 8}, ...]

        Returns:
            ret: 返回当前权重的账号信息 account_arr 中的一个账号信息
        """
        total = 0
        for item in weight_data:
            # 权重求和
            total += item['weight']

        # 在 0 与权重和之间获取一个随机数
        ra = random.uniform(0, total)
        curr_sum = 0
        ret = None
        for key, data in enumerate(weight_data):
            # 在遍历中，累加当前权重值
            curr_sum += data['weight']
            # 当随机数 <= 当前权重和时，返回权重 key
            if ra <= curr_sum:
                ret = data
                break
        return ret

    @classmethod
    def if_dict_meet_min_limit(cls, dict_config: dict, key_list: list) -> bool:
        """
        判断 dict_config 是否满足 key_list 中的 key 值限定
        Args:
            dict_config: 需要判断的参数
            key_list: dict_config 中需要包含的 key 值列表，示例为：['proxy', 'username', 'password']

        Returns:
            1). 是否满足 key 值限制
        """
        if any([not dict_config, not isinstance(dict_config, dict)]):
            return False

        # 理想中的 dict_config 参数为 dict，且其 key 要有且只有 len(key_list) 个
        ideal_keys_list = [x for x in list(dict_config.keys()) if x in key_list]
        # 如果未配置 dict_config 为 dict，且其 key 不是 key_list 中的这几个值时返回 False
        if len(ideal_keys_list) != len(key_list):
            return False
        return True

    @classmethod
    def get_items_by_keys(cls, dict_config: dict, key_list: list) -> Union[dict, bool]:
        """
        获取 dict_config 中的含有 key_list 的 key 的字段
        Args:
            dict_config: 需要处理的参数
            key_list: 需要取的 key 值列表

        Returns:
            1). 取值后的 dict，或不满足请求的 False 值
        """
        # 先要满足最先限定，先要满足这个条件
        if not cls.if_dict_meet_min_limit(dict_config=dict_config, key_list=key_list):
            return False

        return {k: dict_config[k] for k in key_list}

    @classmethod
    def get_items_except_keys(cls, dict_config, key_list: list) -> dict:
        """
        获取 dict_config 中的不含有 key_list 的 key 的字段
        Args:
            dict_config: 需要处理的参数
            key_list: 需要排除的 key 值列表

        Returns:
            1). dict_config 排除 key_list 中的键值后的值
        """

        # 或者这么写
        """
        for key in key_list:
            dict(dict_config).pop(key, None)
        return dict_config
        """
        return {k: dict_config[k] for k in dict_config if k not in key_list}

    @classmethod
    def create_database(cls, pymysql_dict_config: dict):
        """
        创建数据库
        Args:
            pymysql_dict_config: pymysql 的数据库连接配置 dict

        Returns:
            None
        """
        # 判断 pymysql_dict_config 是否满足最少的 key 值
        judge_pymysql_dict_config = cls.if_dict_meet_min_limit(
            dict_config=pymysql_dict_config,
            key_list=["host", "port", "user", "password", "charset"]
        )
        if not judge_pymysql_dict_config:
            raise Exception("创建数据库时的 pymysql 连接参数不满足条件，可能多了 database 参数，或者少了某些参数！")

        pymysql_dict_config_tmp = copy.deepcopy(pymysql_dict_config)
        if "database" in pymysql_dict_config_tmp.keys():
            del pymysql_dict_config_tmp["database"]
        conn = pymysql.connect(**pymysql_dict_config_tmp)
        cursor = conn.cursor()
        cursor.execute(f'''CREATE DATABASE `{pymysql_dict_config["database"]}` character set {pymysql_dict_config["charset"]};''')
        conn.close()
        logger.info(f'''创建数据库 {pymysql_dict_config["database"]} 成功，类型是：{pymysql_dict_config["charset"]}!''')

    @classmethod
    def dict_keys_to_lower(cls, deal_dict: dict) -> dict:
        """
        将 dict 中 str 类型的 key 值变成小写
        Args:
            deal_dict: 需要处理的 dict

        Returns:
            1).处理后的 dict 值
        """
        str_key_to_lower_dict = {k.lower(): v for k, v in deal_dict.items() if isinstance(k, str)}
        not_str_key_dict = {k: v for k, v in deal_dict.items() if not isinstance(k, str)}
        str_key_to_lower_dict.update(not_str_key_dict)
        return str_key_to_lower_dict

    @classmethod
    def dict_keys_to_upper(cls, deal_dict: dict) -> dict:
        """
        将 dict 中 str 类型的 key 值变成大写
        Args:
            deal_dict: 需要处理的 dict

        Returns:
            1).处理后的 dict 值
        """
        # 找出 str 类型的 key 字段数据，并将其大写
        str_key_to_upper_dict = {k.upper(): v for k, v in deal_dict.items() if isinstance(k, str)}
        # 找出非 str 类型的数据
        not_str_key_dict = {k: v for k, v in deal_dict.items() if not isinstance(k, str)}
        # 将大写处理的字典加上非 str 类型的 key 字段数据
        str_key_to_upper_dict.update(not_str_key_dict)
        return str_key_to_upper_dict

    @classmethod
    def get_consul_conf(cls, settings: dict) -> dict:
        """
        获取项目中的 consul 配置，且要根据项目整体情况来取出满足最少要求的 consul 配置
        Args:
            settings: scrapy 的 settings 信息

        Returns:
            consul_conf_dict_min: 满足要求的最少要求的 consul 配置
        """
        consul_conf_dict = settings.get('CONSUL_CONFIG')
        consul_conf_dict_lowered = cls.dict_keys_to_lower(consul_conf_dict)
        # 取最少需要配置的值，consul 一般情况下最少需要 host, port 和 token 共三个值
        consul_conf_dict_min = cls.get_items_by_keys(
            dict_config=consul_conf_dict_lowered,
            key_list=["host", "port", "token"]
        )
        if not consul_conf_dict_min:
            raise Exception(f"consul 配置：{consul_conf_dict} 不满足最小参数配置要求！")

        # 添加 key_values 的值，如果没有设置，则默认取全局配置中的 ENV 值
        if not consul_conf_dict_lowered.get("key_values", None):
            consul_conf_dict_min["key_values"] = settings.get("ENV")
        else:
            consul_conf_dict_min["key_values"] = consul_conf_dict_lowered["key_values"]
        consul_conf_dict_min["group"] = consul_conf_dict_lowered["group"]
        return consul_conf_dict_min

    @classmethod
    def judge_str_is_json(cls, judge_str: str) -> bool:
        """
        判断字符串是否为 json 格式
        Args:
            judge_str: 需要判断的字符串

        Returns:
            1）.是否为 json 格式
        """
        if not isinstance(judge_str, str):
            return False

        try:
            json.loads(judge_str)
        except Exception as e:
            return False
        else:
            return True

    @staticmethod
    def get_ck_dict_from_headers(headers_ck_str: str) -> dict:
        """
        从 headers 中的 ck str 格式转化为 dict 格式
        Args:
            headers_ck_str: request headers ck 的 str 格式

        Returns:
            1). 转化 dict 格式后的 ck
        """
        # 也可以这样写，但不推荐
        # dict(line.split("=", 1) for line in headers_ck_str.split("; "))
        return {x.split("=", 1)[0].strip(): x.split("=", 1)[1].strip() for x in headers_ck_str.split(";")}

    @staticmethod
    def get_req_dict_from_scrapy(req_body_data_str: str) -> dict:
        """
        将 scrapy 请求中的 body 对象转为 dict 格式
        Args:
            req_body_data_str: scrapy 中的 body 参数

        Returns:
            1). 转化 dict 格式后的 body
        """
        return {x.split("=", 1)[0]: x.split("=", 1)[1] for x in req_body_data_str.split("&")}

    @staticmethod
    def get_array_dimension(array: list) -> int:
        """
        获取 array 的维度
        Args:
            array: 数组

        Returns:
            1).层级数
        """
        return len(np.array(array).shape)

    @classmethod
    def get_array_depth(cls, array: list) -> int:
        """
        获取 array 的最大层级，深度
        Args:
            array: 数组

        Returns:
            1).最大层级，深度
        """

        """1 + max(map(depthCount,x)) if x and isinstance(x,list) else 0"""
        # 先判断是否为数组类型的元素
        judge_array = isinstance(array, (frozenset, list, set, tuple,))
        return int(judge_array) and len(array) and 1 + max(map(cls.get_array_depth, array))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@File    :  Oss.py
@Time    :  2022/7/22 13:44
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  None
"""
import oss2
import requests
from retrying import retry
from ayugespidertools.common.Encryption import EncryptOperation
# 去除 requests 的警告信息
import warnings
warnings.filterwarnings('ignore')


__all__ = [
    'AliOssBase',
]


class AliOssBase(object):
    """
    oss 所需要的基本方法
    """
    def __init__(self, OssAccessKeyId, OssAccessKeySecret, Endpoint, examplebucket, operateDoc):
        """
        初始化 auth，bucket 等信息
        注：阿里云账号AccessKey拥有所有API的访问权限，风险很高。强烈建议您创建并使用RAM用户进行API访问或日常运维，请登录RAM控制台创建RAM用户
        Args:
            OssAccessKeyId: 阿里云账号 AccessKey
            OssAccessKeySecret: 阿里云账号 AccessKey 对应的秘钥
            Endpoint: 填写 Bucket 所在地域对应的 Endpoint。
                以华东1（杭州）为例，Endpoint 填写为 https://oss-cn-hangzhou.aliyuncs.com（注意二级域名等问题）
            examplebucket: 填写 Bucket 名称
        """
        self.Endpoint = Endpoint
        self.operateDoc = operateDoc
        self.examplebucket = examplebucket
        self.auth = oss2.Auth(OssAccessKeyId, OssAccessKeySecret)
        self.bucket = oss2.Bucket(self.auth, '{}/'.format(self.Endpoint), self.examplebucket)
        self.headers = {'Connection': 'close', }

    def delete_oss(self, del_logo_url: str):
        """
        删除单个文件: 以下代码用于删除 examplebucket 中的 exampleobject.txt 文件
        Args:
            del_logo_url: 需要参数的阿里云链接全路径 url

        Returns:
            None
        """
        try:
            self.bucket.delete_object('{}/{}'.format(
                self.operateDoc,
                del_logo_url.replace('{}/{}/'.format(self.Endpoint, self.operateDoc), ''))
            )
        except oss2.exceptions.NoSuchKey as e:
            raise Exception('delete_oss error: status={0}, request_id={1}'.format(e.status, e.request_id))

    def old_logo_status(self, old_logo_url: str) -> bool:
        """
        判断 old_logo_url 是否可访问
        Args:
            old_logo_url: 需要判断的链接参数

        Returns:
            1): 链接是否有效
        """
        r = requests.get(old_logo_url, headers=self.headers, verify=False)
        if r.status_code == 200:
            return True
        return False

    @retry(stop_max_attempt_number=3, retry_on_result=lambda x: x[0] is False)
    def put_oss(self, put_bytes_or_url: str, file_name, file_format: str, file_name_md5: bool = False) -> (bool, str):
        """
        上传单个文件的 bytes 内容
        Args:
            file_format: 需要上传的文件格式，比如：jpg, png, wav 等
            put_bytes_or_url: 需要上传的文件 bytes 内容或链接
            file_name: 需要上传的文件的名称
            file_name_md5: 文件名称是否需要 md5 处理

        Returns:
            1): 上传状态，是否成功
            2): 上传成功至 Oss 的目标 url
        """
        input_file_name = EncryptOperation.md5(file_name) if file_name_md5 else file_name
        if all([put_bytes_or_url.startswith("http"), type(put_bytes_or_url) != bytes]):
            put_bytes_or_url = requests.get(url=put_bytes_or_url, headers=self.headers, verify=False).content

        try:
            self.bucket.put_object('{}/{}.{}'.format(self.operateDoc, input_file_name, file_format), put_bytes_or_url)
        except Exception as e:
            return False, ''
        return True, input_file_name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@File    :  format_data.py
@Time    :  2022/7/8 10:33
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  用于数据处理相关方法
"""
import re
import time
import datetime
import html2text
from typing import Optional
from urllib.parse import urljoin
from w3lib.html import remove_tags
from w3lib.html import replace_entities


__all__ = [
    "DataHandle",
]


class DataHandle(object):
    """数据处理相关方法"""

    @staticmethod
    def get_full_url(domain_name: str, deal_url: str) -> str:
        """
        根据域名 domain_name 拼接 deal_url 来获得完整链接
        Args:
            domain_name: 域名链接
            deal_url: 需要拼接的 url

        Returns:
            full_url: 拼接完整的链接
        """
        full_url = urljoin(domain_name, deal_url)
        return full_url

    @staticmethod
    def click_point_deal(decimal: float, decimal_places=2) -> float:
        """
        将小数 decimal 保留小数点后 decimal_places 位，结果四舍五入
        Args:
            decimal: 需要处理的小数
            decimal_places: 需要保留的小数点位数

        Returns:
            decimal(float): 四舍五入后的小数点
        """
        # 先拼接需要保留的位数
        decimal_deal = "%.{}f".format(decimal_places)
        return float(decimal_deal % float(decimal))

    @staticmethod
    def judge_utc_time(local_time: str) -> bool:
        """
        判断 local_time 是否为 utc 格式的时间
        Args:
            local_time: 需要判断的时间参数，比如：Thu Jul 21 17:59:44 2022 或 Fri, 22 Jul 2022 01:43:06 +0800 等等

        Returns:
            1): 是否为 utc 格式的数据
        """
        pattern = re.compile(r"""mon|tues|wed|thu|fri|sat|sun""")
        res = pattern.findall(local_time.lower())
        if res:
            return True
        return False

    @staticmethod
    def judge_include_letter(local_time: str) -> bool:
        pattern = re.compile(r"""[a-zA-Z]""")
        res = pattern.findall(local_time.lower())
        if res:
            return True
        return False

    @staticmethod
    def _get_format_t(
        date_style: str = "",
        date_is_full: bool = False,
        specific_date_conn: str = " ",
        hms_conn: str = ":"
    ) -> str:
        """
        将需要格式化的数据用 date_style 标识来拼接起来，如果 date_is_full 为 True 时，则需要补齐"时分秒"位
        Args:
            date_style: 将格式化符拼接时需要的标识，比如：-
            date_is_full: 是否需要完整的时间格式化（即是否需要补齐"时分秒"单位）
            specific_date_conn: 年月日与时分秒拼接的方式
            hms_conn: 时分秒拼接的方式

        Returns:
            finally_t_format: 最终的格式化拼接结果，比如：%Y-%m-%d %H-%M-%S
        """
        # 年月日
        _y_m_d = ["%Y", "%m", "%d"]
        # 时分秒
        _h_m_s = ["%H", "%M", "%S"]

        _y_m_d_format = date_style.join(_y_m_d)
        # 时分秒大都为":"连接
        _h_m_s_format = hms_conn.join(_h_m_s)

        if date_is_full:
            finally_t_format = specific_date_conn.join([_y_m_d_format, _h_m_s_format])
            return finally_t_format
        return _y_m_d_format

    @staticmethod
    def _time_format(date_str) -> str:
        """
        判断时间是什么格式的，比如：xxxx-xx-xx 或 xxxx.xx.xx
        Args:
            date_str: 需要判断格式的时间

        Returns:
            1): 时间格式的标识，比如：-
        """
        format_style_list = ['-', '.', '/', '年']
        for format_style in format_style_list:
            if format_style in date_str:
                return format_style
        return ""

    @classmethod
    def normal_to_stamp(
        cls,
        normal_time: str,
        _format_t: str = None,
        date_is_full: bool = True,
        specific_date_conn: str = " ",
        hms_conn: str = ":"
    ) -> int:
        """
        将网页正常时间转为时间戳
        Args:
            normal_time: 需要转换的时间
            _format_t: 时间格式化符，默认不填。除非在英文时间的参数出错时可指定 _format_t 的值
            date_is_full: 是否包含"时分秒"单位的完整格式
            specific_date_conn: 年月日与时分秒拼接的方式
            hms_conn: 时分秒拼接的方式

        Returns:
            stamp: 返回的时间戳
        """
        # 判断 normal_time 是否是特殊模式
        is_utc_time = cls.judge_utc_time(normal_time)
        is_letter_time = cls.judge_include_letter(normal_time)
        # stamp = None
        if any([is_utc_time, is_letter_time]):
            # 如果不传递时间的格式符参数
            if not _format_t:
                if "," in normal_time:
                    _format_t = "%a, %d %b %Y %H:%M:%S +0800"
                else:
                    _format_t = "%a %b %d %H:%M:%S %Y"
            standard_time = datetime.datetime.strptime(normal_time, _format_t)
            stamp = time.mktime(time.strptime(str(standard_time), cls._get_format_t('-', True)))

        else:
            # 先判断正常时间的格式
            date_style = cls._time_format(normal_time)
            # standard_time_format = _get_format_t(date_style, date_is_full).replace('%m', '%b').replace('-', '/')
            standard_time_format = cls._get_format_t(date_style, date_is_full, specific_date_conn, hms_conn)
            stamp = time.mktime(time.strptime(normal_time, standard_time_format))
        return int(stamp)

    @staticmethod
    def timestamp_to_normal(timestamp_data: int) -> str:
        """
        将时间戳转为正常时间 xxxx-xx-xx xx:xx:xx 的格式
        Args:
            timestamp_data: 需要处理的时间格式

        Returns:
            ori_handled_time: 转换后的时间结果
        """
        if len(str(timestamp_data)) > 10:
            timestamp_data = int(str(timestamp_data[:-3]))

        time_array = time.localtime(int(timestamp_data))
        ori_handled_time = time.strftime("%Y-%m-%d %H:%M:%S", time_array)
        return ori_handled_time

    @staticmethod
    def remove_all_tags(func):
        """
        去除所有标签
        """
        def inner(*args, **kwargs):
            func_res = func(*args, **kwargs)
            func_res = remove_tags(func_res)
            return func_res
        return inner

    @staticmethod
    def normal_display(func):
        """
        去除掉网页的注释(将网页中的特殊字符的源码改成正常显示)
        """
        def inner(*args, **kwargs):
            func_res = func(*args, **kwargs)
            func_res = replace_entities(func_res)
            return func_res
        return inner

    @staticmethod
    def simple_deal_for_extract(func):
        """
        将 xpath, css 或 json 提取的数据做简单处理；提取的数据若非数组数据，则统一返回字符类型
        """
        def inner(*args, **kwargs):
            func_res = func(*args, **kwargs)
            if type(func_res) in [str, int, float, bool]:
                # 这里可添加一些通用的数据处理
                return str(func_res).strip()
            else:
                return func_res
        return inner

    @classmethod
    def _extract_table_rule(cls, html_txt: str, h_obj: Optional[html2text.HTML2Text] = None):
        """
        根据 html2text 来处理 html 中的 table 表格内容
        Args:
            html_txt: 网页内容
            h_obj: html2text 对象句柄

        Returns:
            1).转换后的结果
        """
        if not h_obj:
            h_obj = html2text.HTML2Text()

        h_obj.protect_links = True
        h_obj.body_width = 0
        h_obj.ignore_links = True
        h_obj.bypass_tables = False
        content = h_obj.handle(html_txt)
        return content

    @classmethod
    def extract_html_to_md(cls, html_txt: str, h_obj: Optional[html2text.HTML2Text] = None) -> str:
        """
        将 html 内容转化为 markdown 内容

        更多、更详细的配置参数请查看文档内容: https://github.com/Alir3z4/html2text/blob/master/html2text/cli.py
        Args:
            html_txt: 网页内容（一般是带标签的内容）
            h_obj: html2text 对象句柄

        Returns:
            1).转换后的结果
        """
        if not h_obj:
            h_obj = html2text.HTML2Text()

        # 以下步骤先处理通用部分，先不处理 table 内容
        h_obj.body_width = 0
        # h_obj.protect_links = True
        h_obj.bypass_tables = True
        # h_obj.unicode_snob = True
        h_obj.ignore_links = True
        content = h_obj.handle(html_txt)

        # 先处理表格 <table> 标签中格式
        table_pattern = re.compile(r"(<table.*?</table>)", flags=re.S)
        table_res_list = table_pattern.findall(content)
        if table_res_list:
            for table_res in table_res_list:
                table_deal = cls._extract_table_rule(html_txt=table_res, h_obj=h_obj)
                # 将 table 转换时去除多余的换行和空白字符等内容
                table_deal = re.sub(r"\|\s*\n", "| ", table_deal, flags=re.S)
                table_deal = re.sub(r"\n\s*\|", "| ", table_deal, flags=re.S)
                table_deal = re.sub(r"\n\s*\n", "\n\n", table_deal, flags=re.S)
                table_deal = re.sub(r"\r\n", "\n", table_deal, flags=re.S)
                table_deal = re.sub(r"[\n]{1,}", "\n", table_deal, flags=re.S)

                content = re.sub(r"(<table.*?</table>)", table_deal + '\n', content, 1, re.S)

        # TODO: 可以添加上通用处理的装饰器
        # 将三个以上换行改为两个换行
        content = re.sub(r"[\n]{3,}", "\n\n", content)
        content = content.strip()
        return content



"""
@File    :  Items.py
@Time    :  2022/7/15 9:38
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  None
"""
from scrapy.item import Item, Field
from dataclasses import dataclass, field


__all__ = [
    "ScrapyClassicItem",
    "MysqlDataItem",
    "MongoDataItem",
    "OldMongoDataItem",
]


class ScrapyClassicItem(Item):
    """
    scrapy 经典 item 示例
    """
    # 用于存放所有字段信息
    alldata = Field()
    # 用于存放存储的表名
    table = Field()
    item_mode = Field()


@dataclass
class BaseItem:
    """
    用于构建 scrapy item 的基本结构，所有需要存储的表对应的结构都放在 alldata 中
    """
    alldata: dict = field(default=None)
    table: str = field(default=None)


@dataclass
class MysqlDataItem(BaseItem):
    """
    这个是 Scrapy item 的 Mysql 的存储结构
    """
    item_mode: str = field(default="Mysql")


@dataclass
class MongoDataItem(BaseItem):
    """
    这个是 Scrapy item 的 MongoDB 的存储结构
    这个 mongo_update_rule 字段是用于 Mongo 存储时作查询使用
    """
    item_mode: str = field(default="MongoDB")
    mongo_update_rule: dict = field(default=None)


@dataclass
class OldMongoDataItem:
    """
    这个是用于 Scrapy item 旧式写法风格存储结构中的默认字段
    这个 mongo_update_rule 字段是用于 Mongo 存储时作查询使用
    """
    item_mode: str = field(default="MongoDB")
    mongo_update_rule: dict = field(default=None)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@File    :  confing.py
@Time    :  2022/7/12 11:22
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  None
"""
import os
import configparser
from environs import Env
from loguru import logger
from os.path import dirname, abspath, join


__all__ = [
    "NormalConfig",
    "logger",
]


env = Env()
env.read_env()


class NormalConfig(object):
    """
    用于存放此项目的通用配置，其实大部分就是用于测试文件使用
    """

    # 项目根目录及其它所需目录
    CONFIG_DIR = dirname(abspath(__file__))
    ROOT_DIR = os.path.dirname(CONFIG_DIR)
    LOG_DIR = join(CONFIG_DIR, "logs")
    DOC_DIR = join(CONFIG_DIR, "docs")
    COMMON_DIR = join(CONFIG_DIR, "common")
    VIT_DIR = join(CONFIG_DIR, "VIT")

    # 从应用配置管理获取配置，是否开启(选择 consul 方式，nacos 也是可以的，以后可能添加从 nacos 中获取配置的功能)
    APP_CONF_MANAGE = env.bool("APP_CONF_MANAGE", True)

    """以下全为本库测试文件所依赖的配置，真实项目中是不需要的。若要调试本库的测试文件，请打开 run_test_debugger 和修改 VIT 文件夹下的 .conf 文件"""
    RUN_TEST_DEBUGGER = env.bool("RUN_TEST_DEBUGGER", False)
    if RUN_TEST_DEBUGGER:
        # 加载秘钥等配置信息
        config_parse = configparser.ConfigParser()
        config_parse.read(f"{VIT_DIR}/.conf", encoding="utf-8")
        # 测试环境中各种配置信息，已脱敏，请自行配置 VIT 的 .conf 文件后测试
        mysql_config = config_parse["DEV_MYSQL"]
        mongodb_config = config_parse["DEV_MONGODB"]
        oss_config = config_parse["DEV_ALI_OSS"]
        # 企业微信机器人 key
        WWXRobot_key = config_parse["WWXRobot"]["key"]

        """配置文件中的 key 值都要大写，但传入程序中的 key 值都要小写，比较复杂的值推荐使用蛇形命名法"""
        # 测试 Mysql 数据库配置
        PYMYSQL_CONFIG = {
            "host": mysql_config["HOST"],
            "port": int(mysql_config["PORT"]),
            "user": mysql_config["USER"],
            "password": mysql_config["PASSWORD"],
            "database": mysql_config["DATABASE"],
            "charset": mysql_config["CHARSET"],
            # "cursorclass": pymysql.cursors.Cursor,
        }

        # 测试 MongoDB 数据库配置
        MONGODB_CONFIG = {
            "host": mongodb_config["HOST"],
            "port": int(mongodb_config["PORT"]),
            "user": mongodb_config["USER"],
            "password": mongodb_config["PASSWORD"],
            "authsource": mongodb_config["AUTHSOURCE"],
            "database": mongodb_config["DATABASE"],
        }

        # 测试 MongoDB 数据库配置
        MONGODB_CONN_URI = config_parse["DEV_MONGODB_URI"]["CONN_URI"]

        # 读取 Oss 配置信息，已脱敏，请自行配置后测试
        OSS_CONFIG = {
            "OssAccessKeyId": oss_config["OSSACCESSKEYID"],
            "OssAccessKeySecret": oss_config["OSSACCESSKEYSECRET"],
            "Endpoint": oss_config["ENDPOINT"],
            "examplebucket": oss_config["EXAMPLEBUCKET"],
            "operateDoc": oss_config["OPERATEDOC"],
        }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@File    :  MultiPlexing.py
@Time    :  2022/7/12 16:53
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  这是完全复制的 scrapy 的 cmdline 内容，用于使用 commands 中的命令
"""
import sys
import os
import argparse
import cProfile
import inspect
import pkg_resources
import ayugespidertools
from scrapy.crawler import CrawlerProcess
from ayugespidertools.commands.version import AyuCommand
from scrapy.commands import ScrapyCommand, ScrapyHelpFormatter
from scrapy.exceptions import UsageError
from scrapy.utils.misc import walk_modules
from scrapy.utils.project import inside_project, get_project_settings
from scrapy.utils.python import garbage_collect


class ScrapyArgumentParser(argparse.ArgumentParser):
    def _parse_optional(self, arg_string):
        # if starts with -: it means that is a parameter not a argument
        if arg_string[:2] == '-:':
            return None

        return super()._parse_optional(arg_string)


def _iter_command_classes(module_name):
    # TODO: add `name` attribute to commands and and merge this function with
    # scrapy.utils.spider.iter_spider_classes
    for module in walk_modules(module_name):
        for obj in vars(module).values():
            if (
                inspect.isclass(obj)
                and issubclass(obj, ScrapyCommand)
                and obj.__module__ == module.__name__
                and not obj == ScrapyCommand
            ):
                yield obj


def _get_commands_from_module(module, inproject):
    d = {}
    for cmd in _iter_command_classes(module):
        if inproject or not cmd.requires_project:
            cmdname = cmd.__module__.split('.')[-1]
            d[cmdname] = cmd()
    return d


def _get_commands_from_entry_points(inproject, group='ayugespidertools.commands'):
    cmds = {}
    for entry_point in pkg_resources.iter_entry_points(group):
        obj = entry_point.load()
        if inspect.isclass(obj):
            cmds[entry_point.name] = obj()
        else:
            raise Exception(f"Invalid entry point {entry_point.name}")
    return cmds


def _get_commands_dict(settings, inproject):
    cmds = _get_commands_from_module('ayugespidertools.commands', inproject)
    cmds.update(_get_commands_from_entry_points(inproject))
    cmds_module = settings['COMMANDS_MODULE']
    if cmds_module:
        cmds.update(_get_commands_from_module(cmds_module, inproject))
    return cmds


def _pop_command_name(argv):
    i = 0
    for arg in argv[1:]:
        if not arg.startswith('-'):
            del argv[i]
            return arg
        i += 1


def _print_header(settings, inproject):
    version = AyuCommand().run(args=None, opts=None)
    if inproject:
        print(f"AyugeSpiderTools {version} - project: {settings['BOT_NAME']}\n")
    else:
        print(f"AyugeSpiderTools {version} - no active project\n")


def _print_commands(settings, inproject):
    _print_header(settings, inproject)
    print("Usage:")
    print("  ayugespidertools <command> [options] [args]\n")
    print("Available commands:")
    cmds = _get_commands_dict(settings, inproject)
    for cmdname, cmdclass in sorted(cmds.items()):
        print(f"  {cmdname:<13} {cmdclass.short_desc()}")
    if not inproject:
        print()
        print("  [ more ]      More commands available when run from project directory")
    print()
    print('Use "ayugespidertools <command> -h" to see more info about a command')


def _print_unknown_command(settings, cmdname, inproject):
    _print_header(settings, inproject)
    print(f"Unknown command: {cmdname}\n")
    print('Use "ayugespidertools" to see available commands')


def _run_print_help(parser, func, *a, **kw):
    try:
        func(*a, **kw)
    except UsageError as e:
        if str(e):
            parser.error(str(e))
        if e.print_help:
            parser.print_help()
        sys.exit(2)


def execute(argv=None, settings=None):
    if argv is None:
        argv = sys.argv

    if settings is None:
        settings = get_project_settings()
        # set EDITOR from environment if available
        try:
            editor = os.environ['EDITOR']
        except KeyError:
            pass
        else:
            settings['EDITOR'] = editor

    inproject = inside_project()
    cmds = _get_commands_dict(settings, inproject)
    cmdname = _pop_command_name(argv)
    if not cmdname:
        _print_commands(settings, inproject)
        sys.exit(0)
    elif cmdname not in cmds:
        _print_unknown_command(settings, cmdname, inproject)
        sys.exit(2)

    cmd = cmds[cmdname]
    parser = ScrapyArgumentParser(formatter_class=ScrapyHelpFormatter,
                                  usage=f"ayugespidertools {cmdname} {cmd.syntax()}",
                                  conflict_handler='resolve',
                                  description=cmd.long_desc())
    settings.setdict(cmd.default_settings, priority='command')
    cmd.settings = settings
    cmd.add_options(parser)
    opts, args = parser.parse_known_args(args=argv[1:])
    _run_print_help(parser, cmd.process_options, args, opts)

    cmd.crawler_process = CrawlerProcess(settings)
    _run_print_help(parser, _run_command, cmd, args, opts)
    sys.exit(cmd.exitcode)


def _run_command(cmd, args, opts):
    if opts.profile:
        _run_command_profiled(cmd, args, opts)
    else:
        cmd.run(args, opts)


def _run_command_profiled(cmd, args, opts):
    if opts.profile:
        sys.stderr.write(f"ayugespidertools: writing cProfile stats to {opts.profile!r}\n")
    loc = locals()
    p = cProfile.Profile()
    p.runctx('cmd.run(args, opts)', globals(), loc)
    if opts.profile:
        p.dump_stats(opts.profile)


if __name__ == '__main__':
    try:
        execute()
    finally:
        # Twisted prints errors in DebugInfo.__del__, but PyPy does not run gc.collect() on exit:
        # http://doc.pypy.org/en/latest/cpython_differences.html
        # ?highlight=gc.collect#differences-related-to-garbage-collection-strategies
        garbage_collect()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@File    :  ProcessManager.py
@Time    :  2022/7/15 9:38
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  None
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@File    :  AyugeSpider.py
@Time    :  2022/7/29 17:33
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  None
"""
import time
import threading
from typing import Union
from scrapy.spiders import Spider
from sqlalchemy import create_engine
from scrapy.spiders import CrawlSpider
from ayugespidertools.common.Utils import ToolsForAyu
from ayugespidertools.common.MultiPlexing import ReuseOperation


__all__ = [
    "AyuSpider",
]


class AyuMixin(Spider):
    """
    防止 AyuSpider 类的 Spider 和 CrawlSpider 出现重复继承的问题
    """

    def parse(self, response, **kwargs):
        super(AyuMixin, self).parse(response, **kwargs)

    def __init__(self, *args, **kwargs):
        super(AyuMixin, self).__init__(*args, **kwargs)


class MySqlEngineClass:
    """
    mysql 链接句柄单例模式
    """
    _instance_lock = threading.Lock()

    def __init__(self, engine_url):
        self.engine = create_engine(engine_url, pool_pre_ping=True, pool_recycle=3600 * 7)

    def __new__(cls, *args, **kwargs):
        if not hasattr(MySqlEngineClass, "_instance"):
            with cls._instance_lock:
                if not hasattr(MySqlEngineClass, "_instance"):
                    MySqlEngineClass._instance = object.__new__(cls)

        return MySqlEngineClass._instance


class AyuSpider(AyuMixin, CrawlSpider):
    """
    用于初始配置 scrapy 的各种 setting 的值及 spider 全局变量等
    """
    # 自定义 common 设置
    custom_common_settings = {
        "ROBOTSTXT_OBEY": False,
        "TELNETCONSOLE_ENABLED": False,
        "RETRY_TIMES": 10,
        "DEPTH_PRIORITY": -1,
        "USER_AGENT": "Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3100.0 Safari/537.36",
        "ITEM_PIPELINES": {
            "ayugespidertools.Pipelines.AyuMysqlPipeline": 300,
        },
        "ENV": "dev",
    }

    # 自定义 Debug 设置
    custom_debug_settings = {
        # 日志等级
        "LOG_LEVEL": "DEBUG",
        "ROBOTSTXT_OBEY": False,
        # 超时
        "DOWNLOAD_TIMEOUT": 30,
        # 重试次数
        "RETRY_TIMES": 50,
        # 禁用所有重定向
        "REDIRECT_ENABLED": False,
        # 后进先出，深度优先
        "DEPTH_PRIORITY": -1,
        # 环境
        "ENV": "test",
        # 数据库表枚举
        "DATA_ENUM": None,
        # 是否记录程序的采集情况基本信息到 Mysql 数据库
        "RECORD_LOG_TO_MYSQL": False,
    }

    # 自定义 product 设置
    custom_product_settings = {
        # 日志等级
        "LOG_LEVEL": "ERROR",
        "ROBOTSTXT_OBEY": False,
        # 超时
        "DOWNLOAD_TIMEOUT": 30,
        # 重试次数
        "RETRY_TIMES": 50,
        # 禁用所有重定向
        "REDIRECT_ENABLED": False,
        # 后进先出，深度优先
        "DEPTH_PRIORITY": -1,
        # 环境
        "ENV": "prod",
        # 数据库表枚举
        "DATA_ENUM": None,
        # 是否记录程序的采集情况基本信息到 Mysql 数据库
        "RECORD_LOG_TO_MYSQL": False,
    }

    # 开始采集时间
    SPIDER_TIME = time.strftime("%Y-%m-%d", time.localtime())
    # 是否启用 Debug 参数，默认激活 custom_common_settings
    settings_type = "common"
    # 脚本信息
    project_content = ""
    # 自定义数据库表枚举
    custom_table_enum = None
    # 数据库引擎开关
    mysql_engine_enabled = False

    def __init__(self, *args, **kwargs):
        super(AyuSpider, self).__init__(*args, **kwargs)
        self.mysql_engine = None

    @classmethod
    def update_settings(cls, settings):
        custom_table_enum = getattr(cls, "custom_table_enum", None)
        settings_type = getattr(cls, "settings_type", "common")
        inner_settings = {}
        if settings_type == "common":
            inner_settings = getattr(cls, "custom_common_settings", {})
        elif settings_type == "debug":
            inner_settings = getattr(cls, "custom_debug_settings", {})
        elif settings_type == "product":
            inner_settings = getattr(cls, "custom_product_settings", {})

        if custom_table_enum:
            inner_settings["DATA_ENUM"] = custom_table_enum

        # 内置配置 inner_settings 优先级介于 project 和 spider 之间, 即优先级顺序：settings.py < inner_settings < custom_settings
        settings.setdict(inner_settings, priority="project")
        settings.setdict(cls.custom_settings or {}, priority="spider")

    @classmethod
    def get_mysql_config(cls, settings) -> Union[dict, None]:
        """
        根据环境获取相应的 Mysql 数据库配置，获取其他自定义配置
        Args:
            settings: crawler settings 配置信息

        Returns:
            1). Mysql 数据库链接配置
        """
        # 自定义 mysql 链接配置
        local_mysql_config = settings.get("LOCAL_MYSQL_CONFIG", {})
        # 是否开启应用配置管理
        app_conf_manage = settings.get("APP_CONF_MANAGE", False)
        # 优先从本地取配置，先判断是否配置两个中的任意一个
        if all([not local_mysql_config.get("DATABASE"), not app_conf_manage]):
            return None

        # 1). 从本地 local_mysql_config 的参数中取值
        if ReuseOperation.if_dict_meet_min_limit(
            dict_config=local_mysql_config,
            key_list=["HOST", "PORT", "USER", "PASSWORD", "CHARSET", "DATABASE"]
        ):
            mysql_conf_temp = {
                "host": local_mysql_config.get("HOST"),
                "port": local_mysql_config.get("PORT"),
                "user": local_mysql_config.get("USER"),
                "password": local_mysql_config.get("PASSWORD"),
                "database": local_mysql_config.get("DATABASE"),
                "charset": local_mysql_config.get("CHARSET") or "utf8mb4"
            }
            return mysql_conf_temp

        # 2). 本地没有配置，再从 consul 中获取应用配置
        if app_conf_manage:
            consul_conf_dict_min = ReuseOperation.get_consul_conf(settings=settings)
            return ToolsForAyu.get_mysql_conf_by_consul(**consul_conf_dict_min)

    @classmethod
    def get_mongodb_config(cls, settings) -> Union[dict, None]:
        """
        根据环境获取相应的 mongoDB 数据库配置，获取其他自定义配置
        Args:
            settings: crawler settings 配置信息

        Returns:
            1). MongoDB 数据库链接配置
        """
        # 自定义 mysql 链接配置
        local_mongodb_conf = settings.get("LOCAL_MONGODB_CONFIG", {})
        # 是否开启应用配置管理
        app_conf_manage = settings.get("APP_CONF_MANAGE", False)
        # 优先从本地取配置
        if all([not local_mongodb_conf.get("DATABASE"), not app_conf_manage]):
            return None

        # 1). 从本地 local_mongo_conf 的参数中取值
        if ReuseOperation.if_dict_meet_min_limit(
            dict_config=local_mongodb_conf,
            key_list=["HOST", "PORT", "USER", "PASSWORD", "DATABASE"]
        ):
            mongodb_conf_temp = {
                "host": local_mongodb_conf.get("HOST"),
                "port": local_mongodb_conf.get("PORT"),
                "user": local_mongodb_conf.get("USER"),
                "password": local_mongodb_conf.get("PASSWORD"),
                "database": local_mongodb_conf.get("DATABASE"),
            }
            return mongodb_conf_temp

        # 2). 本地没有配置，再从 consul 中获取应用配置
        if app_conf_manage:
            consul_conf_dict_min = ReuseOperation.get_consul_conf(settings=settings)
            return ToolsForAyu.get_mongodb_conf_by_consul(**consul_conf_dict_min)

    @classmethod
    def from_crawler(cls, crawler, *args, **kwargs):
        spider = super(AyuSpider, cls).from_crawler(crawler, *args, **kwargs)
        # 1).先配置 Mysql 的相关信息，如果存在 Mysql 配置，则把 mysql_conf 添加到 spider 上
        mysql_conf = cls.get_mysql_config(crawler.settings)
        # 如果配置了 Mysql 信息
        if mysql_conf:
            spider.logger.info("项目中配置了 mysql_config 信息")
            spider.mysql_config = mysql_conf
            spider.stats = crawler.stats

            # 如果打开了 mysql_engine_enabled 参数(用于 spiders 中数据入库前去重查询)
            if cls.mysql_engine_enabled:
                mysql_url = "mysql+pymysql://{}:{}@{}:{}/{}?charset={}".format(
                    mysql_conf.get("user"),
                    mysql_conf.get("password"),
                    mysql_conf.get("host"),
                    mysql_conf.get("port"),
                    mysql_conf.get("database"),
                    mysql_conf.get("charset")
                )
                spider.mysql_engine = MySqlEngineClass(engine_url=mysql_url).engine

        # 2).配置 MongoDB 的相关信息，如果存在 MongoDB 配置，则把 mongodb_conf 添加到 spider 上
        mongodb_conf = cls.get_mongodb_config(crawler.settings)
        # 如果配置了 MongoDB 信息
        if mongodb_conf:
            spider.logger.info("项目中配置了 mongodb_config 信息")
            spider.mongodb_conf = mongodb_conf
        return spider



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@File    :  YiDunGap.py
@Time    :  2022/7/12 10:36
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  None
"""
import cv2
from ayugespidertools.common.MultiPlexing import ReuseOperation


__all__ = [
    "YiDunGetGap",
]


class YiDunGetGap(object):
    """
    易盾获取滑块缺口距离的相关方法，也可能适配于其它平台
    """

    @classmethod
    def clear_white(cls, img):
        """
        清除图片的空白区域，这里主要清除滑块的空白
        Args:
            img: 待处理的图片

        Returns:
            img1: 清楚图片空白区域的图片
        """
        rows, cols, channel = img.shape
        min_x = 255
        min_y = 255
        max_x = 0
        max_y = 0
        for x in range(1, rows):
            for y in range(1, cols):
                t = set(img[x, y])
                if len(t) >= 2:
                    if x <= min_x:
                        min_x = x
                    elif x >= max_x:
                        max_x = x

                    if y <= min_y:
                        min_y = y
                    elif y >= max_y:
                        max_y = y

        img1 = img[min_x:max_x, min_y: max_y]
        return img1

    @classmethod
    def template_match(cls, tpl, target, out: str = None) -> int:
        """
        模板匹配找出滑块缺口的距离
        Args:
            tpl: 缺口图片
            target: 背景图片
            out: 展示图片的存储全路径，会将绘制的图片保存至此

        Returns:
            tl[0]: 滑块缺口的距离
        """
        th, tw = tpl.shape[:2]
        result = cv2.matchTemplate(target, tpl, cv2.TM_CCOEFF_NORMED)
        # 寻找矩阵(一维数组当作向量,用Mat定义) 中最小值和最大值的位置
        min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(result)
        tl = max_loc

        # 是否展示标注后的图片
        if out:
            br = (tl[0] + tw, tl[1] + th)
            # 绘制矩形边框，将匹配区域标注出来
            # target：目标图像
            # tl：矩形定点
            # br：矩形的宽高
            # (0,0,255)：矩形边框颜色
            # 1：矩形边框大小
            cv2.rectangle(target, tl, br, (0, 0, 255), 2)
            cv2.imwrite(out, target)
        return tl[0]

    @classmethod
    def image_edge_detection(cls, img):
        """
        图像边缘检测处理，识别图片边缘
        Args:
            img: 需要处理的图片，用于边缘检测使用

        Returns:
            处理后的图片
        """
        edges = cv2.Canny(img, 100, 200)
        return edges

    @classmethod
    def discern(cls, slide, bg, out: str = None):
        """
        识别滑块缺口方法
        Args:
            slide: 滑块图，可以是全路径图片，也可以是图片的 bytes 数据
            bg: 带缺口的背景图，可以是全路径图片，也可以是图片的 bytes 数据
            out: 绘制图展示的存储地址，参数格式为图片的全路径

        Returns:
            x: 滑块缺口横坐标
        """
        # 先用 opencv 读取图片数据
        slide_cv, bg_cv = ReuseOperation.read_image_data(slide, bg)

        # 清除图片的空白区域
        img1 = cls.clear_white(slide_cv)
        img1 = cv2.cvtColor(img1, cv2.COLOR_RGB2GRAY)
        # 图像边缘检测处理
        slide = cls.image_edge_detection(img1)
        back = cls.image_edge_detection(bg_cv)

        slide_pic = cv2.cvtColor(slide, cv2.COLOR_GRAY2RGB)
        back_pic = cv2.cvtColor(back, cv2.COLOR_GRAY2RGB)
        # 输出横坐标, 即滑块在图片上的位置
        x = cls.template_match(slide_pic, back_pic, out)
        return x



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@File    :  VerificationCode.py
@Time    :  2022/7/11 14:01
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  None
"""
import cv2
import math
import random
import numpy as np
from ayugespidertools.common.YiDunGap import YiDunGetGap


__all__ = [
    'match_img_get_distance',
    'get_selenium_tracks',
    'get_yidun_tracks',
    'get_yidun_gap',
    'get_normal_track',
]


def match_img_get_distance(target, template):
    """
    网易易盾（dun163）滑块坐标定位方法，也使用于其它平台滑块验证码的情况
    Args:
        target: 1) 带缺口的背景图，全路径; 2) 或者是背景图的 bytes 数据
        template: 1） 滑块图，全路径; 2）或者是滑块图的 bytes 数据

    Returns:
        loc[1][0]: 滑块位置坐标
    """
    # 先判断传入的参数是 str 图片的全路径信息还是 bytes 类型的图片信息
    if any([type(target) != str, type(target) == bytes]):
        target_buf = np.frombuffer(target, np.uint8)
        template_buf = np.frombuffer(template, np.uint8)

        target_cv = cv2.imdecode(target_buf, cv2.IMREAD_ANYCOLOR)
        template_cv = cv2.imdecode(template_buf, cv2.IMREAD_ANYCOLOR)

    else:
        # 读取图片，读进来直接是 BGR 格式数据格式在 0~255
        target_cv = cv2.imread(target)
        # 0 表示采用黑白的方式读取图片
        template_cv = cv2.imread(template, 0)

        # cv2.cvtColor(p1,p2) 是颜色空间转换函数，p1是需要转换的图片，p2是转换成何种格式。
        # cv2.COLOR_BGR2GRAY 将BGR格式转换成灰度图片，发现转换后并不是通常意义上的黑白图片。
        # 灰度图片并不是指常规意义上的黑白图片，只用看是不是无符号八位整型（unit8）,单通道即可判断。
        target_cv = cv2.cvtColor(target_cv, cv2.COLOR_BGR2GRAY)

    run = 1
    # 这里的返回值根据不同版本的 open-cv 其返回结果会有不同的个数
    # w, h, z = template_cv.shape[::-1]
    # 在背景图里面查找滑块图的位置
    res = cv2.matchTemplate(target_cv, template_cv, cv2.TM_CCOEFF_NORMED)
    # 使用二分法查找阈值的精确值
    L = 0
    R = 1
    loc = None
    while run < 20:
        run += 1
        threshold = (R + L) / 2
        if threshold < 0:
            # 逻辑走到这里，则说明出错了，并未找出目标位置
            return None

        # 匹配程度大于百分之 threshold 的坐标 x, y
        loc = np.where(res >= threshold)
        if len(loc[1]) > 1:
            L += (R - L) / 2
        elif len(loc[1]) == 1:
            # 找到目标区域起点x坐标为：loc[1][0]
            break
        elif len(loc[1]) < 1:
            R -= (R - L) / 2

    # 返回 x 坐标
    return loc[1][0]


def get_selenium_tracks(distance):
    """
    最简陋的根据缺口距离获取轨迹的方法，以供 selenium 使用
    Args:
        distance: 滑块缺口的距离

    Returns:
        tracks_dict:
            forward_tracks: 往右划的轨迹数组
            back_tracks: 回退（往左划）的轨迹数组
    """
    distance += 20
    v = 0
    t = 0.2
    forward_tracks = []
    current = 0
    mid = distance * 3 / 5
    while current < distance:
        if current < mid:
            a = 2
        else:
            a = -3
        s = v * t + 0.5 * a * (t ** 2)
        v = v + a * t
        current += s
        forward_tracks.append(round(s))

    back_tracks = [-3, -3, -2, -2, -2, -2, -2, -1, -1, -1]
    return {'forward_tracks': forward_tracks, 'back_tracks': back_tracks}


def get_yidun_tracks(distance):
    """
    网易易盾轨迹生成方法（不能直接用极验的轨迹生成和其它如（加减速、匀加速或者先加速后减速，甚至划过再划回）简单生成的方法），生成的轨迹数组中首个位移最好要大于等于 3
    Args:
        distance: 滑块缺口的距离

    Returns:
        xyt: 网易易盾轨迹数组
    """
    t_list = [random.randint(50, 160)]
    x_list = [random.randint(5, 11)]
    y_list = list()
    # 生成x坐标轨迹, 生成t坐标轨迹
    for j in range(1, distance):
        x_list.append(x_list[j - 1] + random.randint(2, 4))
        if x_list[j] > distance:
            break

    diff = x_list[-1] - distance
    for j in range(diff):
        x_list.append(x_list[-1] + random.randint(-2, -1))
        if x_list[-1] <= distance:
            x_list[-1] = distance
            break

    length = len(x_list)
    # 生成y坐标轨迹
    for i in range(1, length + 1):
        if i < int(length * 0.4):
            y_list.append(0)
        elif i < int(length * 0.65):
            y_list.append(-1)
        elif i < int(length * 0.77):
            y_list.append(-2)
        elif i < int(length * 0.95):
            y_list.append(-3)
        else:
            y_list.append(-4)
        t_list.append(t_list[i - 1] + random.randint(20, 80))

    # 生成t的坐标
    xyt = list(zip(x_list, y_list, t_list))
    for j in range(length):
        xyt[j] = list(xyt[j])
    return xyt


def get_yidun_gap(slide_img_path, bg_img_path, out_img_path: str = False) -> int:
    """
    获取易盾滑块缺口距离
    Args:
        slide_img_path: 滑块的图片
        bg_img_path: 带缺口的背景图片
        out_img_path: 输出图片

    Returns:
        gap_distance: 易盾滑块缺口距离
    """
    gap_distance = YiDunGetGap.discern(slide_img_path, bg_img_path, out_img_path)
    return gap_distance


def get_normal_track(space):
    """
    通用的根据滑块距离获取轨迹数组方法
    Args:
        space: 滑块缺口距离

    Returns:
        tracks_list: 生成的轨迹数组
    """
    x = [0, 0]
    y = [0, 0, 0]
    z = [0]
    # x
    count = np.linspace(-math.pi / 2, math.pi / 2, random.randrange(20, 30))
    func = list(map(math.sin, count))
    nx = [i + 1 for i in func]
    add = random.randrange(10, 15)
    sadd = space + add
    x.extend(list(map(lambda x: x * (sadd / 2), nx)))
    # x.extend(np.linspace(sadd, space, 4 if add > 12 else 3))
    x.extend(np.linspace(sadd, space, 3 if add > 12 else 2))
    x = [math.floor(i) for i in x]
    # y
    for i in range(len(x) - 2):
        if y[-1] < 30:
            y.append(y[-1] + random.choice([0, 0, 1, 1, 2, 2, 1, 2, 0, 0, 3, 3]))
        else:
            y.append(y[-1] + random.choice([0, 0, -1, -1, -2, -2, -1, -2, 0, 0, -3, -3]))
    # z
    for i in range(len(x) - 1):
        # z.append((z[-1] // 100 * 100) + 100 + random.choice([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 2]))
        z.append((z[-1] // 100 * 100) + 100 + random.choice([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 3]))

    tracks_list = list(map(list, zip(x, y, z)))
    tracks_list = [x for x in tracks_list if x[0] > 0]
    return tracks_list



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@File    :  MongoClient.py
@Time    :  2022/7/22 15:43
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  None
"""
from gridfs import *
from typing import Optional
from typing import List, Dict
from pymongo import MongoClient


__all__ = [
    'MongoDbBase',
]


class MongoDbBase(object):
    """
    mongodb 数据库的相关操作（此功能暂时为残废状态，请参考 pymilk 库中的实现）
    """
    def __init__(
        self,
        user: str,
        password: str,
        host: str,
        port: int,
        authsource: str = "admin",
        database: str = None,
        connect_style: str = None
    ) -> None:
        """
        初始化 mongo 连接句柄
        Args:
            user: 用户名
            password: 用户对应的密码
            host: mongoDB 链接需要的 host
            port: mongoDB 链接需要的端口
            authsource: mongoDB 身份验证需要的数据库名称，默认为 admin
            database: mongoDB 链接需要的数据库
            connect_style: mongoDB 的链接方式，参数选择有：
                            1). uri: uri 方式；
                            2). key: 关键字变量方式；
                            3). auth: authenticate 或 admin 认证方式;
                            默认为 uri 方式，但其实对外使用还是只传 user, password ... connect_style 的形式
        """
        # uri 方式，默认使用此方式连接
        if any([not connect_style, connect_style in ["uri", "U"]]):
            uri = "mongodb://%s:%s@%s:%s/?authSource=%s&authMechanism=SCRAM-SHA-1" % (
                user, password, host, port, authsource)
            self.conn = MongoClient(uri)

        # 关键字变量方式
        elif connect_style in ["key", "K"]:
            self.conn = MongoClient(
                host=host,
                port=port,
                username=user,
                password=password,
                authSource=authsource,
                authMechanism='SCRAM-SHA-1'
            )

        # authenticate 或 admin 认证方式(pymongo 3.9 及以下版本使用)，不推荐此方法
        elif connect_style in ["auth", "A"]:
            self.conn = MongoClient(host, port)
            # 连接 admin 数据库，账号密码认证(其实这里也可以使用 uri 的 auth 认证方式)
            db = self.conn.admin
            db.authenticate(user, password)

        else:
            raise Exception("你指定错误了 connect_style 的 mongo 链接类型，请正确输入！")

        if database:
            self.db = self.init_db(database)

    def get_state(self):
        """
        获取 mongoDB 链接状态
        Returns:
            1). bool: 链接是否正常
        """
        return all([self.conn is not None, self.db is not None])

    def init_db(self, database: str):
        """
        指定链接的数据库为 database
        Args:
            database: 链接的目标数据库

        Returns:
            1). connect: 数据库链接
        """
        return self.conn[database]

    def insert_one(self, collection: str, data: dict) -> str:
        """
        插入一条数据
        Args:
            collection: 集合名称
            data: 插入的数据

        Returns:
            inserted_id: 成功返回的 id
        """
        if self.get_state():
            ret = self.db[collection].insert_one(data)
            return ret.inserted_id
        return ""

    def insert_many(self, collection: str, data: List[Dict]):
        """
        批量插入
        Args:
            collection: 集合名称
            data: 插入的数据数组

        Returns:
            inserted_ids: list: 成功返回的 ids
        """
        if self.get_state():
            ret = self.db[collection].insert_many(data)
            return ret.inserted_ids
        return ""

    def update(self, collection, data):
        """
        更新
        Args:
            collection: 集合名称
            data: 更新的数据，{key:[old_data,new_data]}

        Returns:
            modified_count: 更新影响的个数
        """
        data_filter = {}
        data_revised = {}
        for key in data.keys():
            data_filter[key] = data[key][0]
            data_revised[key] = data[key][1]
        if self.get_state():
            return self.db[collection].update_many(data_filter, {"$set": data_revised}).modified_count
        return 0

    def find(self, collection, condition, column: Optional[dict] = None):
        """
        查询
        Args:
            collection: 集合名称
            condition: 查询条件
            column:

        Returns:
            1). 查询条件的搜索结果
        """
        try:
            if self.get_state():
                if column is None:
                    return self.db[collection].find(condition)
                else:
                    return self.db[collection].find(condition, column)
            else:
                return None
        except Exception:
            return "查询数据格式有误"

    def t_update(self, collection, s_key, s_value, update_data):
        # self.db.get_collection(collection).update_one({s_key: s_value}, {"$set": update_data})
        return self.db[collection].update_one({s_key: s_value}, {"$set": {"groupProps": update_data}}).modified_count

    def update_normal(self, collection, s_key, s_value, data_style, update_data):
        return self.db[collection].update_one({s_key: s_value}, {"$set": {data_style: update_data}}).modified_count

    def update_super(self, collection: str, select_dict: dict, set_dict: dict):
        """
        更新
        Args:
            collection: 需要更新的集合
            select_dict: 更新的条件
            set_dict: 更新的内容

        Returns:
            1). modified_count: int: 更新影响的数量
        """
        return self.db[collection].update_one(select_dict, {"$set": set_dict}).modified_count

    # 删除
    def delete(self, collection, condition):
        if self.get_state():
            return self.db[collection].delete_many(filter=condition).deleted_count
        return 0

    # 上传数据
    def upload_file(self, file_name, collection, content_type, file_data, metadata):
        gridfs_col = GridFS(self.db, collection)
        # with open(file_name, 'rb') as file_r:
        #     file_data = file_r.read()
        #     file_ = gridfs_col.put(data=file_data, content_type=content_type, filename=file_name, metadata=metadata)
        #     # return file_

        file_ = gridfs_col.put(data=file_data, content_type=content_type, filename=file_name, metadata=metadata)
        return file_

    def getFileMd5(self, _id, collection):
        gridfs_col = GridFS(self.db, collection)
        gf = gridfs_col.get(_id)
        md5 = gf.md5
        _id = gf._id
        return {'_id': _id, 'md5': md5}

    def upload(self, file_name, _id, content_type, collection, file_data):
        """
        上传文件
        Args:
            file_name: 上传至 mongoDB 的 GridFS 存储桶里的文件名
            _id: 唯一 id，雪花 id，用来标识此上传任务和图片的唯一
            content_type: 上传文件的类型，示例：image/jpeg
            collection: 存储至 GridFS 存储桶名称
            file_data: 文件的 bytes 内容

        Returns:
            gridfs_id: 上传至 GridFS 上的文件 ID 标识
            image_id: 上传至 GridFS 上的文件 MD5 标识
        """
        metadata = {
            "_contentType": content_type,
            "isThumb": 'true',
            "targetId": _id,
            "_class": "com.ccr.dc.admin.mongo.MongoFsMetaData"
        }

        gridfs_col = GridFS(self.db, collection)
        # 当存储桶中不存在此文件，才需要上传
        if not gridfs_col.exists(filename=file_name):
            gridfs_id = gridfs_col.put(data=file_data, content_type=content_type, filename=file_name, metadata=metadata)
            md5 = self.getFileMd5(gridfs_id, 'fs')['md5']
            image_id = f"/file/find/{str(gridfs_id)}/{md5}"
            return gridfs_id, image_id

        # 否则，只需返回文件的 id 等标识即可
        res = gridfs_col.find_one({"filename": file_name})
        return res._id, f"/file/find/{str(res._id)}/{res.md5}"

    def close_mongodb(self):
        """手动关闭连接"""
        self.conn.close()

    def __del__(self):
        self.conn.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@File    :  ImgOperation.py
@Time    :  2022/7/12 15:14
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  None
"""
import cv2
import requests
from PIL import Image
from ayugespidertools.config import NormalConfig
from ayugespidertools.common.MultiPlexing import ReuseOperation


__all__ = [
    'Picture',
]


class Picture(object):
    """
    对验证码图片的一些操作
    """

    @classmethod
    def get_captcha(cls, url: str):
        """
        下载完美滑块的图片，并将缺口图和滑块在一起的图片切割
        Args:
            url: 完美滑块的滑块图片链接

        Returns:
            None
        """
        session = requests.Session()
        session.headers = {
            'authority': 'captchas-1251008858.file.myqcloud.com',
            'pragma': 'no-cache',
            'cache-control': 'no-cache',
            'sec-ch-ua': '" Not;A Brand";v="99", "Google Chrome";v="91", "Chromium";v="91"',
            'sec-ch-ua-mobile': '?0',
            'upgrade-insecure-requests': '1',
            'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.77 Safari/537.36',
            'accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9',
            'sec-fetch-site': 'none',
            'sec-fetch-mode': 'navigate',
            'sec-fetch-user': '?1',
            'sec-fetch-dest': 'document',
            'accept-language': 'zh-CN,zh;q=0.9,en;q=0.8',
        }
        text = session.get(url).content
        with open(f'{NormalConfig.DOC_DIR}/captcha.png', 'wb') as f:
            f.write(text)

        # captcha = Image.new('RGB', (50, 120))  # 新建空白图片
        img = Image.open(f'{NormalConfig.DOC_DIR}/captcha.png')  # 实例化原始图片Image对象

        # 切割滑块验证码图片，将背景图和滑块图分开
        captcha = img.crop((260, 0, 325, 120 - 4))  # (left, upper, right, lower)
        # captcha = img.crop((274, 46, 300, 120 - 20))  # (left, upper, right, lower)
        captcha = captcha.convert('RGBA')
        captcha.save(f'{NormalConfig.DOC_DIR}/captcha_slide.png')

    @classmethod
    def convert_index_to_offset(cls, index):
        """
        获取每张小图的偏移量
        Args:
            index: 当前小块图片

        Returns:
            当前小图在空白图片的坐标
        """

        # 代码注释
        '''
        # 图片大小为：334 159， 22 为 334 / 15 得到，58 是 159 /
        if index < 15:  # 完整的验证码图片是由30个小图片组合而成，共2行15列
            return (index * 22, 0)
        else:
            i = index - 15
            return (i * 22, 58)  # 每张小图的大小为22*58
        '''
        # 完整的验证码图片是由 40 个小图片组合而成，共 2 行 15 列
        if index < 20:
            return index * 13, 0
        else:
            i = index - 20
            # 每张小图的大小为 22 * 58
            return i * 13, 58

    @classmethod
    def convert_css_to_offset(cls, off):
        """
        获取每张小图的坐标，供抠图时使用
        Args:
            off: 根据 css backgound-position 中获取的每张小图的坐标

        Returns:
            (int(off[0]), int(off[1]), int(off[0]) + 22, int(off[1]) + 58): 每张小图对应的坐标
        """
        # (left, upper)o ----- o
        #              |       |
        #              o ----- o(right, lower)
        return int(off[0]), int(off[1]), int(off[0]) + 22, int(off[1]) + 58

    @classmethod
    def recombine_captcha(cls, offset_list):
        """
        图片重组: 完美世界网站的图片重组方法
        Args:
            offset_list: 坐标列表

        Returns:
            None
        """
        captcha = Image.new('RGB', (13 * 20, 60 * 2 - 4))  # 新建空白图片
        img = Image.open(f'{NormalConfig.DOC_DIR}/captcha.png')  # 实例化原始图片Image对象

        # 切割滑块验证码图片，将背景图和滑块图分开
        captcha_de = img.crop((0, 0, 260, 120 - 4))  # (left, upper, right, lower)
        captcha_de = captcha_de.convert('RGBA')
        captcha_de.save(f'{NormalConfig.DOC_DIR}/captcha.png')

        for i, off in enumerate(offset_list):
            box = Picture.convert_css_to_offset(off)  # 根据css backgound-position获取每张小图的坐标
            regoin = img.crop(box)  # 抠图
            offset = Picture.convert_index_to_offset(i)  # 获取当前小图在空白图片的坐标
            captcha.paste(regoin, offset)  # 根据当前坐标将小图粘贴到空白图片
        captcha.save(f'{NormalConfig.DOC_DIR}/regoin.jpg')

    @classmethod
    def reset_pic(cls, slide_data):
        """
        完美世界滑块验证码重组方法，具体位置根据 background-position 搜索定位
        Args:
            slide_data: 完美滑块重组所需的数组

        Returns:
            true_pic_list: 真实图片的顺序坐标
        """
        c = 260
        d = 120
        l = 20
        s = 9
        a = 61
        true_pic_list = []
        for curr_data in slide_data:
            curr_position_list = []
            if curr_data < l:
                curr_position_list.append(int(c / l * curr_data))
                curr_position_list.append(0)

            else:
                curr_position_list.append(int(c / l * (curr_data % l)))
                curr_position_list.append(60)

            true_pic_list.append(curr_position_list)
        return true_pic_list

    @classmethod
    def find_pic(cls, target, template):
        """
        找出图像中最佳匹配位置
        Args:
            target: 目标（背景图）
            template: 模板（需要找到的图）

        Returns:
            value[2:][0][0]: 返回最佳匹配及对应的坐标
            value[2:][1][0]: 返回最差匹配及对应的坐标
        """
        target_rgb = cv2.imread(target)
        target_gray = cv2.cvtColor(target_rgb, cv2.COLOR_BGR2GRAY)
        template_rgb = cv2.imread(template, 0)
        res = cv2.matchTemplate(target_gray, template_rgb, cv2.TM_CCOEFF_NORMED)
        value = cv2.minMaxLoc(res)
        return value[2:][0][0], value[2:][1][0]

    @classmethod
    def identify_gap(cls, bg, tp, out: str = None) -> int:
        """
        通过背景图片和缺口图片识别出滑块距离
        Args:
            bg: 背景图片，可以是图片的全路径，也可以是图片的 bytes 内容
            tp: 缺口（滑块）图片，可以是图片的全路径，也可以是图片的 bytes 内容
            out: 输出图片路径，示例：doc/test.jpg

        Returns:
            tl[0]: 滑块缺口距离
        """
        # 先读使用 opencv 读取图片数据
        bg_cv, tp_cv = ReuseOperation.read_image_data(bg, tp)

        # 识别图片边缘
        bg_edge = cv2.Canny(bg_cv, 100, 200)
        tp_edge = cv2.Canny(tp_cv, 100, 200)

        # 转换图片格式
        bg_pic = cv2.cvtColor(bg_edge, cv2.COLOR_GRAY2RGB)
        tp_pic = cv2.cvtColor(tp_edge, cv2.COLOR_GRAY2RGB)

        # 缺口匹配
        res = cv2.matchTemplate(bg_pic, tp_pic, cv2.TM_CCOEFF_NORMED)
        min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res)  # 寻找最优匹配
        tl = max_loc  # 左上角点的坐标

        # 是否要输出绘制图像
        if out:
            # 绘制方框
            th, tw = tp_pic.shape[:2]
            br = (tl[0] + tw, tl[1] + th)  # 右下角点的坐标
            cv2.rectangle(bg_cv, tl, br, (0, 0, 255), 2)  # 绘制矩形
            cv2.imwrite(out, bg_cv)  # 保存在本地

        # 返回缺口的X坐标
        return tl[0]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@File    :  Encryption.py
@Time    :  2022/7/22 14:42
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  None
"""
import base64
import hashlib


__all__ = [
    'EncryptOperation',
]


class EncryptOperation(object):
    """
    普通加密方法
    """

    @classmethod
    def md5(cls, encrypt_data: str) -> str:
        """
        md5 处理方法
        Args:
            encrypt_data: 需要 md5 处理的参数

        Returns:
            1): md5 处理后的参数
        """
        hl = hashlib.md5()
        hl.update(encrypt_data.encode(encoding='utf-8'))
        return hl.hexdigest()

    @classmethod
    def base64_encode(cls, encode_data: str, url_safe: bool = False) -> str:
        """
        base64 编码
        Args:
            encode_data: 需要 base64 编码的参数
            url_safe: 标准的 Base64 编码后可能出现字符 + 和 /，在 url 中就不能直接作为参数。是否要处理此情况

        Returns:
            1). base64 编码后的结果
        """
        if not isinstance(encode_data, bytes):
            encode_data = bytes(encode_data, encoding="utf-8")
        if url_safe:
            return str(base64.urlsafe_b64encode(encode_data), encoding="utf-8")
        return str(base64.b64encode(encode_data), encoding="utf-8")

    @classmethod
    def base64_decode(cls, decode_data: str, url_safe: bool = False) -> str:
        """
        base64 解码
        Args:
            decode_data: 需要 base64 解码的参数
            url_safe: 标准的 Base64 编码后可能出现字符 + 和 /，在 url 中就不能直接作为参数。是否要处理此情况

        Returns:
            1). base64 解码后的结果
        """
        if url_safe:
            return str(base64.urlsafe_b64decode(decode_data), encoding="utf-8")
        return str(base64.b64decode(decode_data), encoding="utf-8")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@File    :  Utils.py
@Time    :  2022/8/5 16:15
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  用于 AyugeSpiderTools 的一些通用方法，主要是用于 scrapy 扩展的一些通用方法
"""
import copy
import json
import requests
import dataclasses
import ayugespidertools.Items
from typing import Union, List
from itemadapter import ItemAdapter
from ayugespidertools.config import logger
from ayugespidertools.FormatData import DataHandle
from ayugespidertools.common.Params import Param
from ayugespidertools.common.Encryption import EncryptOperation
from ayugespidertools.common.MultiPlexing import ReuseOperation


__all__ = [
    "ToolsForAyu",
]


class ToolsForAyu(object):
    """
    这里用于存放框架所依赖的方法
    """

    @classmethod
    def consul_get_all_group(cls, host: str, port: int, token: str) -> list:
        """
        获取 consul 的所有 group 信息
        Args:
            host: host
            port: port
            token: 请求 consul 所需要的 token 值

        Returns:
            1). consul 的所有 group 信息
        """
        curr_consul_headers = copy.deepcopy(Param.consul_headers)
        curr_consul_headers["X-Consul-Token"] = token
        r = requests.get(
            url=f'http://{host}:{port}/v1/kv/?keys&dc=dc1&separator=%2F',
            headers=curr_consul_headers,
            verify=False
        )
        return r.json()

    @classmethod
    def get_kvs_detail_by_consul(cls, host: str, port: int, token: str, key_values: str, group: str = None) -> dict:
        """
        获取 consul 的 key_values 的详细信息
        Args:
            host: consul host
            port: consul port
            token: consul token，最好只要有只读权限的 token 即可
            key_values: 需要获取的 consul key_values 参数值
            group: key_values 所属的 group，如果没有的话不设置此值即可

        Returns:
            conf_value: consul 的 group 下 key_values 的详细信息
        """
        if group:
            consul_url = f"http://{host}:{port}/v1/kv/{group}/{key_values}?dc=dc1"
        else:
            consul_url = f"http://{host}:{port}/v1/kv/{key_values}?dc=dc1"

        curr_consul_headers = copy.deepcopy(Param.consul_headers)
        curr_consul_headers["X-Consul-Token"] = token
        r = requests.get(consul_url, headers=curr_consul_headers, verify=False)
        conf_value = EncryptOperation.base64_decode(decode_data=r.json()[0]["Value"])
        return json.loads(conf_value)

    @classmethod
    def get_mysql_conf_by_consul(cls, host: str, port: int, token: str, key_values: str, group: str = None):
        """
        获取 consul 中的 mysql 配置信息
        Args:
            host: consul host
            port: consul port
            token: consul token，最好只要有只读权限的 token 即可
            key_values: 需要获取的 consul key_values 参数值
            group: key_values 所属的 group，如果没有的话不设置此值即可

        Returns:
            mysql_conf_lowered: consul 应用配置中心中的 Mysql 配置信息（key 值为小写）
        """
        conf_value = cls.get_kvs_detail_by_consul(host, port, token, key_values, group)
        mysql_conf = conf_value["MYSQL"] or conf_value["mysql"]
        mysql_conf_lowered = ReuseOperation.dict_keys_to_lower(mysql_conf)
        return mysql_conf_lowered

    @classmethod
    def get_mongodb_conf_by_consul(cls, host: str, port: int, token: str, key_values: str, group: str = None):
        """
        获取 consul 中的 mysql 配置信息
        Args:
            host: consul host
            port: consul port
            token: consul token，最好只要有只读权限的 token 即可
            key_values: 需要获取的 consul key_values 参数值
            group: key_values 所属的 group，如果没有的话不设置此值即可

        Returns:
            mysql_conf_lowered: consul 应用配置中心中的 Mysql 配置信息（key 值为小写）
        """
        conf_value = cls.get_kvs_detail_by_consul(host, port, token, key_values, group)
        mysql_conf = conf_value["MONGODB"] or conf_value["mongodb"]
        mysql_conf_lowered = ReuseOperation.dict_keys_to_lower(mysql_conf)
        return mysql_conf_lowered

    @classmethod
    @DataHandle.simple_deal_for_extract
    def extract_with_css(cls, response, query: str, get_all: bool = False, return_selector: bool = False):
        """
        使用 scrapy 的 css 提取信息
        Args:
            response: scrapy response 或者是 selector 对象
            query: css 提取的规则
            get_all: 提取模式，默认：False, 提取符合中的一个
            return_selector: 是否返回选择器对象

        Returns:
            1).提取的内容
        """
        if return_selector:
            return response.css(query)
        if get_all:
            return response.css(query).getall()
        else:
            return response.css(query).get(default="").strip()

    @classmethod
    @DataHandle.simple_deal_for_extract
    def extract_with_xpath(cls, response, query: str, get_all: bool = False, return_selector: bool = False):
        """
        使用 scrapy 的 xpath 提取信息
        Args:
            response: scrapy response 或者是 selector 对象
            query: css 提取的规则
            get_all: 提取模式，默认：False, 提取符合中的一个
            return_selector: 是否返回选择器对象

        Returns:
            1).提取的内容
        """
        if return_selector:
            return response.xpath(query)
        if get_all:
            return response.xpath(query).getall()
        else:
            return response.xpath(query).get(default="").strip()

    @classmethod
    @DataHandle.simple_deal_for_extract
    def extract_with_json(cls, json_data: dict, query: Union[str, List[str]]):
        """
        scrapy 中提取 json 数据遇到的情况
        Args:
            json_data: scrapy response 响应内容中的 json 格式数据
            query: json 提取的规则

        Returns:
            1).提取的内容
        """
        # 如果输入的提取规则参数的格式为字符；或者参数格式是个列表，但是只含有一个元素的情况时
        if any([isinstance(query, str), all([isinstance(query, list), len(query) == 1])]):
            if isinstance(query, str):
                return json_data.get(query, "")
            return json_data.get(query[0], "")

        # 循环取值时的处理
        for curr_q in query:

            # 这里循环时不对 json_data 的类型做判断，如果此时 json_data 类型无 get 方法，不处理
            json_data = json_data.get(curr_q, "")
            if not json_data:
                return json_data
        return json_data

    @classmethod
    def extract_with_json_rules(cls, json_data: dict, query_rules: List[Param.Str_Lstr]):
        """
        当提取 json 某个数据时，可以在某些字段中取值，只要返回其中任意一个含有数据的值即可
        Args:
            json_data: scrapy response 响应内容中的 json 格式数据
            query_rules: json 提取的规则列表

        Returns:
            1).提取的内容
        """
        # 先判断层级是否为两层
        depth_num = ReuseOperation.get_array_depth(query_rules)
        if depth_num > 2:
            raise Exception("query_rules 参数错误，请输入深度最多为 2 的参数！")

        for query in query_rules:
            extract_res = cls.extract_with_json(json_data=json_data, query=query)

            # 只要有任意字段存在数据的情况就返回
            if extract_res:
                return extract_res
        return ""

    @staticmethod
    def get_collate_by_charset(mysql_config: dict) -> str:
        """
        根据 mysql 的 charset 获取对应 collate
        Args:
            mysql_config: mysql 连接配置

        Returns:
            collate: 排序规则
        """
        if mysql_config["charset"] == "utf8mb4":
            collate = "utf8mb4_general_ci"
        elif mysql_config["charset"] == "utf8":
            collate = "utf8_general_ci"
        else:
            raise Exception(f"数据库连接时出现未知 charset：{mysql_config['charset']}，若抛错请查看！")
        return collate

    @staticmethod
    def convert_item_to_dict(item) -> dict:
        """
        将 item 结构数据转为 dict 格式（这种转换方法太过啰嗦，已放弃，请使用下面的 convert_items_to_dict 方法）
        Args:
            item: 需要转换的参数

        Returns:
            1). 转换为 dict 格式的结果
        """
        # MongoDataItem 等其它的场景不再转换为 dict，直接使用 dataclass 即可
        if isinstance(item, ayugespidertools.Items.MysqlDataItem):
            return dataclasses.asdict(item)

        elif isinstance(item, dict):
            return item

        elif isinstance(item, ayugespidertools.Items.ScrapyClassicItem):
            return dict(item)

        else:
            logger.warning(f"出现未知 item 格式，item: {item}")
            return dict(item)

    @staticmethod
    def convert_items_to_dict(item) -> ItemAdapter:
        """
        数据容器对象的包装器，提供了一个通用接口以统一的方式处理不同类型的对象，而不管它们的底层实现如何。
        目前支持的类型有：
            1. scrapy.item.Item
            2. dict
            3. dataclass 基础类
            4. attrs 基础类
            5. pydantic 基础类
        Args:
            item: 需要转换的项目，请查看支持类型

        Returns:
            1). 转换的 ItemAdapter 结果，可以通过  obj["params"] 或 obj.get("params") 来取值
        """
        return ItemAdapter(item)

    @staticmethod
    def get_dict_form_scrapy_req_headers(scrapy_headers) -> dict:
        """
        根据 scrapy request 中的 headers 信息转化为正常的 dict 格式
        Args:
            scrapy_headers: scrapy 的 request headers 内容

        Returns:
            req_headers: 转化 dict 后的 headers 内容
        """
        req_headers = dict()
        for b_key, b_value_list in dict(scrapy_headers).items():
            req_headers[str(b_key, encoding="utf-8")] = str(b_value_list[0], encoding="utf-8")
        return req_headers



from os.path import join, exists, abspath
import ayugespidertools
from scrapy.commands.startproject import Command


class AyuCommand(Command):

    def run(self, *args, **kwargs):
        super(AyuCommand, self).run(*args, **kwargs)
        print("Or you can start your first spider with ayugespidertools:")
        print("    ayugespidertools genspider example example.com")

    @property
    def templates_dir(self):
        return join(join(ayugespidertools.__path__[0], 'templates'), 'project')



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
from enum import Enum


class TableEnum(Enum):
    """
    数据库表枚举信息示例
    """

    aritle_list_table = {"value": "article_info_list", "notes": "项目列表信息", "demand_code": "DemoSpider_aritle_list_table_demand_code"}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@File    :  AyuRequest.py
@Time    :  2022/8/30 14:52
@Author  :  Ayuge
@Version :  1.0
@Contact :  ayuge.s@qq.com
@License :  (c)Copyright 2022-2023
@Desc    :  None
"""
import copy
from scrapy import Request, FormRequest
from parsel.selector import create_root_node
from w3lib.html import strip_html5_whitespace
from scrapy.utils.response import get_base_url
from ayugespidertools.common.Params import Param
from scrapy.http.response.text import TextResponse
from scrapy.utils.python import to_bytes, is_listlike
from urllib.parse import urljoin, urlencode, urlsplit, urlunsplit
from lxml.html import FormElement, HtmlElement, HTMLParser, SelectElement
from typing import Callable, Iterable, List, Optional, Tuple, Type, TypeVar, Union


__all__ = [
    "AiohttpRequest",
    "AioFormRequest",
]


FormRequestTypeVar = TypeVar("FormRequestTypeVar", bound="FormRequest")
FormdataType = Optional[Union[dict, List[Tuple[str, str]]]]


class AiohttpRequest(Request):
    """
    为 scrapy 的 Request 对象添加额外的参数
    """

    def __init__(
        self,
        url: str,
        callback: Optional[Callable] = None,
        method: str = "GET",
        headers: Optional[dict] = None,
        body: Optional[Union[bytes, str]] = None,
        cookies: Optional[Union[dict, List[dict]]] = None,
        meta: Optional[dict] = None,
        *args,
        **kwargs
    ) -> None:
        # 用 meta 缓存 scrapy meta 的参数
        meta = copy.deepcopy(meta) or {}
        aiohttp_meta = meta.get("aiohttp_args") or {}

        # TODO: 可添加和修改默认参数的值，后续可以在此完善功能
        self.proxy = aiohttp_meta.get("proxy", {"http:": "http://10.10.10.10:10000"})
        self.timeout = aiohttp_meta.get("timeout", Param.aiohttp_req_timeout)
        self.sleep = aiohttp_meta.get("sleep", None)

        aiohttp_meta = meta.setdefault("aiohttp_args", {})
        aiohttp_meta["proxy"] = self.proxy
        aiohttp_meta["timeout"] = self.timeout
        super(AiohttpRequest, self).__init__(
            url,
            callback,
            method=method,
            headers=headers,
            body=body,
            cookies=cookies,
            meta=meta,
            *args,
            **kwargs
        )


class AioFormRequestBackup(AiohttpRequest, FormRequest):
    """
    这里只是用于测试，后期会删除此部分
    """

    def __init__(
        self,
        url: str,
        callback: Optional[Callable] = None,
        method: str = "GET",
        headers: Optional[dict] = None,
        body: Optional[Union[bytes, str]] = None,
        cookies: Optional[Union[dict, List[dict]]] = None,
        meta: Optional[dict] = None,
        formdata: FormdataType = None,
        *args,
        **kwargs
    ) -> None:
        super().__init__(
            url,
            callback,
            method,
            headers,
            body,
            cookies,
            meta,
            formdata,
            *args,
            **kwargs
        )

    @classmethod
    def from_response(
            cls: Type[FormRequestTypeVar],
            response: TextResponse,
            formname: Optional[str] = None,
            formid: Optional[str] = None,
            formnumber: Optional[int] = 0,
            formdata: FormdataType = None,
            clickdata: Optional[dict] = None,
            dont_click: bool = False,
            formxpath: Optional[str] = None,
            formcss: Optional[str] = None,
            **kwargs,
    ) -> FormRequestTypeVar:
        return super().from_response(
            response,
            formname,
            formid,
            formnumber,
            formdata,
            clickdata,
            dont_click,
            formxpath,
            formcss,
            **kwargs
        )


class AioFormRequest(AiohttpRequest):
    """
    为 scrapy 的 FormRequest 对象添加额外的参数
    (这里直接 copy 的 scrapy 的 FormRequest 内容)
    """

    valid_form_methods = ['GET', 'POST']

    def __init__(self, *args, formdata: FormdataType = None, **kwargs) -> None:
        if formdata and kwargs.get('method') is None:
            kwargs['method'] = 'POST'

        super().__init__(*args, **kwargs)

        if formdata:
            items = formdata.items() if isinstance(formdata, dict) else formdata
            form_query_str = _urlencode(items, self.encoding)
            if self.method == 'POST':
                self.headers.setdefault(b'Content-Type', b'application/x-www-form-urlencoded')
                self._set_body(form_query_str)
            else:
                self._set_url(urlunsplit(urlsplit(self.url)._replace(query=form_query_str)))

    @classmethod
    def from_response(
            cls: Type[FormRequestTypeVar],
            response: TextResponse,
            formname: Optional[str] = None,
            formid: Optional[str] = None,
            formnumber: Optional[int] = 0,
            formdata: FormdataType = None,
            clickdata: Optional[dict] = None,
            dont_click: bool = False,
            formxpath: Optional[str] = None,
            formcss: Optional[str] = None,
            **kwargs,
    ) -> FormRequestTypeVar:
        kwargs.setdefault('encoding', response.encoding)

        if formcss is not None:
            from parsel.csstranslator import HTMLTranslator
            formxpath = HTMLTranslator().css_to_xpath(formcss)

        form = _get_form(response, formname, formid, formnumber, formxpath)
        formdata = _get_inputs(form, formdata, dont_click, clickdata)
        url = _get_form_url(form, kwargs.pop('url', None))

        method = kwargs.pop('method', form.method)
        if method is not None:
            method = method.upper()
            if method not in cls.valid_form_methods:
                method = 'GET'

        return cls(url=url, method=method, formdata=formdata, **kwargs)


def _get_form_url(form: FormElement, url: Optional[str]) -> str:
    if url is None:
        action = form.get('action')
        if action is None:
            return form.base_url
        return urljoin(form.base_url, strip_html5_whitespace(action))
    return urljoin(form.base_url, url)


def _urlencode(seq: Iterable, enc: str) -> str:
    values = [(to_bytes(k, enc), to_bytes(v, enc))
              for k, vs in seq
              for v in (vs if is_listlike(vs) else [vs])]
    return urlencode(values, doseq=True)


def _get_form(
    response: TextResponse,
    formname: Optional[str],
    formid: Optional[str],
    formnumber: Optional[int],
    formxpath: Optional[str],
) -> FormElement:
    """Find the wanted form element within the given response."""
    root = create_root_node(response.text, HTMLParser, base_url=get_base_url(response))
    forms = root.xpath('//form')
    if not forms:
        raise ValueError(f"No <form> element found in {response}")

    if formname is not None:
        f = root.xpath(f'//form[@name="{formname}"]')
        if f:
            return f[0]

    if formid is not None:
        f = root.xpath(f'//form[@id="{formid}"]')
        if f:
            return f[0]

    # Get form element from xpath, if not found, go up
    if formxpath is not None:
        nodes = root.xpath(formxpath)
        if nodes:
            el = nodes[0]
            while True:
                if el.tag == 'form':
                    return el
                el = el.getparent()
                if el is None:
                    break
        raise ValueError(f'No <form> element found with {formxpath}')

    # If we get here, it means that either formname was None or invalid
    if formnumber is not None:
        try:
            form = forms[formnumber]
        except IndexError:
            raise IndexError(f"Form number {formnumber} not found in {response}")
        else:
            return form


def _get_inputs(
    form: FormElement,
    formdata: FormdataType,
    dont_click: bool,
    clickdata: Optional[dict],
) -> List[Tuple[str, str]]:
    """Return a list of key-value pairs for the inputs found in the given form."""
    try:
        formdata_keys = dict(formdata or ()).keys()
    except (ValueError, TypeError):
        raise ValueError('formdata should be a dict or iterable of tuples')

    if not formdata:
        formdata = []
    inputs = form.xpath('descendant::textarea'
                        '|descendant::select'
                        '|descendant::input[not(@type) or @type['
                        ' not(re:test(., "^(?:submit|image|reset)$", "i"))'
                        ' and (../@checked or'
                        '  not(re:test(., "^(?:checkbox|radio)$", "i")))]]',
                        namespaces={"re": "http://exslt.org/regular-expressions"})
    values = [
        (k, '' if v is None else v)
        for k, v in (_value(e) for e in inputs)
        if k and k not in formdata_keys
    ]

    if not dont_click:
        clickable = _get_clickable(clickdata, form)
        if clickable and clickable[0] not in formdata and not clickable[0] is None:
            values.append(clickable)

    if isinstance(formdata, dict):
        formdata = formdata.items()  # type: ignore[assignment]

    values.extend((k, v) for k, v in formdata if v is not None)
    return values


def _value(ele: HtmlElement):
    n = ele.name
    v = ele.value
    if ele.tag == 'select':
        return _select_value(ele, n, v)
    return n, v


def _select_value(ele: SelectElement, n: str, v: str):
    multiple = ele.multiple
    if v is None and not multiple:
        # Match browser behaviour on simple select tag without options selected
        # And for select tags without options
        o = ele.value_options
        return (n, o[0]) if o else (None, None)
    elif v is not None and multiple:
        # This is a workround to bug in lxml fixed 2.3.1
        # fix https://github.com/lxml/lxml/commit/57f49eed82068a20da3db8f1b18ae00c1bab8b12#L1L1139
        selected_options = ele.xpath('.//option[@selected]')
        values = [(o.get('value') or o.text or '').strip() for o in selected_options]
        return n, values
    return n, v


def _get_clickable(clickdata: Optional[dict], form: FormElement) -> Optional[Tuple[str, str]]:
    """
    Returns the clickable element specified in clickdata,
    if the latter is given. If not, it returns the first
    clickable element found
    """
    clickables = list(form.xpath(
        'descendant::input[re:test(@type, "^(submit|image)$", "i")]'
        '|descendant::button[not(@type) or re:test(@type, "^submit$", "i")]',
        namespaces={"re": "http://exslt.org/regular-expressions"}
    ))
    if not clickables:
        return None

    # If we don't have clickdata, we just use the first clickable element
    if clickdata is None:
        el = clickables[0]
        return (el.get('name'), el.get('value') or '')

    # If clickdata is given, we compare it to the clickable elements to find a
    # match. We first look to see if the number is specified in clickdata,
    # because that uniquely identifies the element
    nr = clickdata.get('nr', None)
    if nr is not None:
        try:
            el = list(form.inputs)[nr]
        except IndexError:
            pass
        else:
            return (el.get('name'), el.get('value') or '')

    # We didn't find it, so now we build an XPath expression out of the other
    # arguments, because they can be used as such
    xpath = './/*' + ''.join(f'[@{k}="{v}"]' for k, v in clickdata.items())
    el = form.xpath(xpath)
    if len(el) == 1:
        return (el[0].get('name'), el[0].get('value') or '')
    elif len(el) > 1:
        raise ValueError(f"Multiple elements found ({el!r}) matching the "
                         f"criteria in clickdata: {clickdata!r}")
    else:
        raise ValueError(f'No clickable element matching clickdata: {clickdata!r}')



# Define here the models for your spider middleware
#
# See documentation in:
# https://docs.scrapy.org/en/latest/topics/spider-middleware.html
import json
import time
import base64
import random
import requests
import traceback
import numpy as np
from retrying import retry
from scrapy import signals
from typing import Optional
from threading import Thread
from scrapy.http import HtmlResponse
from scrapy.settings import Settings
from .config import NormalConfig, logger
from .common.MultiPlexing import ReuseOperation
from ayugespidertools.common.Params import Param
from ayugespidertools.common.Utils import ToolsForAyu
from scrapy.utils.project import get_project_settings
from WorkWeixinRobot.work_weixin_robot import WWXRobot
from scrapy.utils.response import response_status_message
from scrapy.downloadermiddlewares.retry import RetryMiddleware


__all__ = [
    "RandomRequestUaMiddleware",
    "DynamicProxyDownloaderMiddleware",
    "ExclusiveProxyDownloaderMiddleware",
    "AbuDynamicProxyDownloaderMiddleware",
    "RequestByRequestsMiddleware",
]


class RandomRequestUaMiddleware(object):
    """
    随机请求头中间件
    """
    def get_random_ua_by_weight(self) -> str:
        """根据权重来获取随机请求头 ua 信息"""
        # 带权重的 ua 列表，将比较常用的 ua 标识的权重设置高一点。这里是根据 fake_useragent 库中的打印信息来规划权重的。
        ua_arr = [
            {"explorer_type": "opera", "weight": 16},
            {"explorer_type": "safari", "weight": 32},
            {"explorer_type": "internetexplorer", "weight": 41},
            {"explorer_type": "firefox", "weight": 124},
            {"explorer_type": "chrome", "weight": 772},
        ]
        curr_ua_type = ReuseOperation.random_weight(weight_data=ua_arr)
        curr_ua_list = Param.fake_useragent_dict[curr_ua_type["explorer_type"]]
        curr_ua = random.choice(curr_ua_list)
        return curr_ua

    @classmethod
    def from_crawler(cls, crawler):
        # This method is used by Scrapy to create your spiders.
        s = cls()
        crawler.signals.connect(s.spider_opened, signal=signals.spider_opened)
        return s

    def spider_opened(self, spider):
        logger.info(f"随机请求头中间件 RandomRequestUaMiddleware 已开启，生效脚本为: %s" % spider.name)
        # spider.logger.info("RandomRequestUaMiddleware opened: %s" % spider.name)

    def process_request(self, request, spider):
        curr_ua = self.get_random_ua_by_weight()
        if curr_ua:
            request.headers.setdefault(b"User-Agent", curr_ua)
            spider.logger.info("RandomRequestUaMiddleware 当前使用的 ua 为: %s" % curr_ua)


class DynamicProxyDownloaderMiddleware(object):
    """
    动态隧道代理中间件
    """
    def __init__(self, settings):
        """
        从 scrapy 配置中取出动态隧道代理的信息
        """
        dynamic_proxy_config = settings.get("DYNAMIC_PROXY_CONFIG", None)
        # 查看动态隧道代理配置是否符合要求
        is_match = ReuseOperation.if_dict_meet_min_limit(
            dict_config=dynamic_proxy_config,
            key_list=["PROXY_URL", "USERNAME", "PASSWORD"]
        )
        if not is_match:
            raise Exception(f"没有配置动态隧道代理，配置示例为：{Param.dynamic_proxy_config_example}")

        self.proxy_url = dynamic_proxy_config["PROXY_URL"]
        self.username = dynamic_proxy_config["USERNAME"]
        self.password = dynamic_proxy_config["PASSWORD"]

    @classmethod
    def from_crawler(cls, crawler):
        # This method is used by Scrapy to create your spiders.
        s = cls(crawler.settings)
        crawler.signals.connect(s.spider_opened, signal=signals.spider_opened)
        return s

    def process_request(self, request, spider):
        # TODO: 根据权重来随机获取一个账号 DYNAMIC_PROXY_CONFIG
        # account = ReuseOperation.random_weight(self.account_arr)
        if request.url.startswith("https://"):
            request.meta['proxy'] = f"https://{self.username}:{self.password}@{self.proxy_url}/"
        elif request.url.startswith("http://"):
            request.meta['proxy'] = f"http://{self.username}:{self.password}@{self.proxy_url}/"
        else:
            spider.logger.info(f"request url: {request.url} error when use proxy middlewares!")

        # 避免因连接复用导致隧道不能切换 IP
        request.headers["Connection"] = "close"
        # 采用 gzip 压缩加速访问
        request.headers["Accept-Encoding"] = "gzip"

    def spider_opened(self, spider):
        logger.info(f"动态隧道代理中间件: DynamicProxyDownloaderMiddleware 已开启，生效脚本为: {spider.name}")


class ExclusiveProxyDownloaderMiddleware(object):
    """
    独享代理中间件
    """

    def __init__(self, exclusive_proxy_config):
        """
        初始化独享代理设置
        Args:
            exclusive_proxy_config: 使用的独享代理的配置信息
        """
        self.proxy = None
        # 查看独享代理配置是否符合要求
        is_match = ReuseOperation.if_dict_meet_min_limit(
            dict_config=exclusive_proxy_config,
            key_list=["PROXY_URL", "USERNAME", "PASSWORD", "PROXY_INDEX"]
        )
        if not is_match:
            raise Exception(f"没有配置独享代理，配置示例为：{Param.exclusive_proxy_config_example}")

        self.proxy_url = exclusive_proxy_config["PROXY_URL"]
        self.username = exclusive_proxy_config["USERNAME"]
        self.password = exclusive_proxy_config["PASSWORD"]
        self.proxy_index = exclusive_proxy_config["PROXY_INDEX"]

    @classmethod
    def from_crawler(cls, crawler):
        # This method is used by Scrapy to create your spiders.
        s = cls(
            exclusive_proxy_config=crawler.settings.get("EXCLUSIVE_PROXY_CONFIG", None)
        )
        crawler.signals.connect(s.spider_opened, signal=signals.spider_opened)
        return s

    def get_proxy_ip(self):
        """获取独享代理接口的索引为 proxy_index 的代理信息"""
        try:
            r = requests.get(self.proxy_url)
            proxy_list = r.json().get("data").get("proxy_list")
            proxy_list.sort()
            if self.proxy_index < len(proxy_list):
                self.proxy = proxy_list[self.proxy_index]
            else:
                raise Exception("独享代理索引超出范围，请确认独享代理服务情况。")

        except Exception:
            raise Exception("获取独享代理是失败，请及时查看。")

    def process_request(self, request, spider):
        if request.url.startswith("https://"):
            request.meta["proxy"] = "https://{}".format(self.proxy)
        elif request.url.startswith("http://"):
            request.meta["proxy"] = "http://{}".format(self.proxy)
        else:
            spider.logger.info(f"request url error: {request.url}")

        proxy_user_pass = self.username + ":" + self.password
        encoded_user_pass = "Basic " + base64.urlsafe_b64encode(bytes(proxy_user_pass, "ascii")).decode("utf8")
        request.headers["Proxy-Authorization"] = encoded_user_pass

    def spider_opened(self, spider):
        self.get_proxy_ip()
        logger.info(f"独享代理中间件: ExclusiveProxyDownloaderMiddleware 已开启，生效脚本为: {spider.name}，当前独享代理为: {self.proxy}")


class SiMiProxyDownloaderMiddleware(RetryMiddleware):
    """
    基于独享代理的自定义动态代理中间件
    """

    def __init__(self):
        self.settings = Settings()
        super(SiMiProxyDownloaderMiddleware, self).__init__(self.settings)
        self.important_error = False
        self.WWX = WWXRobot(key=NormalConfig.WWXRobot_key)

    @classmethod
    def from_crawler(cls, crawler):
        # This method is used by Scrapy to create your spiders.
        s = cls()
        crawler.signals.connect(s.spider_opened, signal=signals.spider_opened)
        crawler.signals.connect(s.spider_closed, signal=signals.spider_closed)
        return s

    @retry(stop_max_attempt_number=Param.retry_num)
    def get_proxy_ip(self, size, isdict: Optional[bool] = None):
        proxy_url = "http://dps.kdlapi.com/api/getdps?orderid={}&num={}&signature={}&format=json".format(
            self.simidaili_config["orderid"], size, self.simidaili_config["signature"]
        )
        if not self.important_error:
            try:
                r = requests.get(proxy_url).json()
            except Exception:
                raise ValueError("快代理请求获取 ip 无响应，请检查是否能够正常访问快代理或者 nacos 的配置是否以及失效！")
        else:
            raise ValueError("ip 获取方式有误，请重构私密代理中间件获取 ip 的模块！")

        if r["code"] == 0:
            try:
                iplist = r.get("data").get("proxy_list")
                now_ipnum = r.get("data").get("today_left_count")
                if iplist and now_ipnum:
                    if (now_ipnum < 100) and (now_ipnum >= 50):
                        self.WWX.send_text("私密代理三级警报：相关部门请注意，ip 数量已经少于 100 个！请及时充值！")
                    elif (now_ipnum < 50) and (now_ipnum >= 10):
                        self.WWX.send_text("私密代理二级警报：相关部门请注意，相关部门请注意, ip 数量已经少 50 个！请及时充值！")
                    elif now_ipnum < 10:
                        self.WWX.send_text("私密代理一级警报：相关部门请注意，，相关部门请注意，相关部门请注意, ip 数量已经少于 10 个！请及时充值！")

                    if isdict:
                        self.proxy_list = iplist
                    else:
                        return iplist
                else:
                    raise ValueError("ip 获取方式有误，请重构私密代理中间件获取 ip 的模块！")
            except:
                self.important_error = True
                traceback.print_exc()
                raise ValueError("ip 获取方式有误，请重构私密代理中间件获取ip的模块！")

        else:
            msg = r.get("msg")
            if msg:
                self.WWX.send_text(f"私密代理特级警报：私密代理出现错误，错误原因是: {msg}")
                raise ValueError(f"获取私密ip失败，原因是：{msg}")
            else:
                self.important_error = True
                raise ValueError("ip 获取方式有误，请重构私密代理中间件获取 ip 的模块！")

    def process_request(self, request, spider):
        # 初始化请求，随机选定一个 ip 进行访问
        current_ip_index = np.argmin(self.v_count)
        current_ip = self.proxy_list[current_ip_index]

        # 统计请求速率
        self.reqnum_count[current_ip_index] += 1.0
        self.v_sum()
        # logger.debugger(f"目前的平均请求速度为：{np.average(self.v_count)}")

        if request.url.startswith("https://"):
            request.meta["proxy"] = "https://{}".format(current_ip)
        elif request.url.startswith("http://"):
            request.meta["proxy"] = "http://{}".format(current_ip)
        else:
            spider.logger.info(f"request url error: {request.url}")

        # request.meta["proxy"] = "http://{}".format(current_ip)
        request.meta["dont_retry"] = True
        # request.meta["ja3"] = True
        proxy_user_pass = self.username + ":" + self.password
        encoded_user_pass = "Basic " + base64.urlsafe_b64encode(bytes(proxy_user_pass, "ascii")).decode("utf8")
        request.headers["Proxy-Authorization"] = encoded_user_pass

    def process_response(self, request, response, spider):
        if response.status in self.retry_http_codes:
            # 如果出现返回异常的状态码则更换ip进行请求
            current_ip_index = np.argmin(self.v_count)
            current_ip = self.proxy_list[current_ip_index]

            # 统计请求速率
            self.reqnum_count[current_ip_index] += 1
            self.v_sum()

            if request.url.startswith("https://"):
                request.meta["proxy"] = "https://{}".format(current_ip)
            elif request.url.startswith("http://"):
                request.meta["proxy"] = "http://{}".format(current_ip)
            else:
                spider.logger.info(f"request url error: {request.url}")
            # request.meta["proxy"] = "http://{}".format(current_ip)
            reason = response_status_message(response.status)
            return self._retry(request, reason, spider) or response
        return response

    def process_exception(self, request, exception, spider):
        if isinstance(exception, self.EXCEPTIONS_TO_RETRY):
            current_ip = request.meta["proxy"].split(""//"")[-1]
            url = "https://dps.kdlapi.com/api/checkdpsvalid?orderid={}&signature={}&proxy={}".format(
                self.simidaili_config["orderid"], self.simidaili_config["signature"],
                current_ip)
            is_exists = requests.get(url).json()
            if not is_exists["data"][current_ip]:
                if current_ip in self.proxy_list:
                    spider.logger.info("更新ip")
                    # 找到相应索引，并删除相应时间统计数组、ip列表以及请求次数统计数组相应索引上的值
                    current_ip_index = self.proxy_list.index(current_ip)
                    self.proxy_list.pop(current_ip_index)
                    self.time_count = np.delete(self.time_count, current_ip_index, axis=0)
                    self.reqnum_count = np.delete(self.reqnum_count, current_ip_index, axis=0)

                    # 如果平均速度大于每个 ip 4 次每秒，或者 ip 池子小于 1 时，则增加两个 ip，否则不予增加。
                    if (np.average(self.v_count) >= 4.5) and (len(self.proxy_list) < 7):
                        for new_ip in self.get_proxy_ip(2):
                            self.proxy_list.append(new_ip)
                            self.time_count = np.append(self.time_count, 0.5)
                            self.reqnum_count = np.append(self.reqnum_count, 0)
                            current_ip = new_ip
                            self.reqnum_count[-1] += 1
                    elif len(self.proxy_list) < self.proxypool_size:
                        new_ip = self.get_proxy_ip(1)[0]
                        self.proxy_list.append(new_ip)
                        self.time_count = np.append(self.time_count, 0.5)
                        self.reqnum_count = np.append(self.reqnum_count, 0)
                        current_ip = new_ip
                        self.reqnum_count[-1] += 1
                    else:
                        current_ip_index = np.argmin(self.v_count)
                        current_ip = self.proxy_list[current_ip_index]
                        self.reqnum_count[current_ip_index] += 1

                    spider.logger.info(f"目前 ip 池存量 ip 数量为：{len(self.proxy_list)}")
                    self.v_sum()
                else:
                    current_ip_index = np.argmin(self.v_count)
                    current_ip = self.proxy_list[current_ip_index]
                    self.reqnum_count[current_ip_index] += 1
                    self.v_sum()

            if request.url.startswith("https://"):
                request.meta["proxy"] = "https://{}".format(current_ip)
            elif request.url.startswith("http://"):
                request.meta["proxy"] = "http://{}".format(current_ip)
            else:
                spider.logger.info(f"request url error: {request.url}")
            # request.meta["proxy"] = "http://{}".format(current_ip)
            return self._retry(request, exception, spider)

    def v_sum(self):
        self.v_count = self.reqnum_count / self.time_count

    def clock(self, spider):
        while not self.end_time:
            time.sleep(0.2)
            self.time_count = np.add(self.time_count, 0.2)

    def spider_opened(self, spider):
        settings = get_project_settings()
        # TODO: 根据 nacos/consul 获取代理配置信息
        # resx = get_nacos_config(env="prod", data_id="KUAI_PROXY_PRIVATE")
        resx = json.dumps({"username": "***", "password": "***"})
        self.simidaili_config = json.loads(resx)
        self.username = self.simidaili_config["username"]
        self.password = self.simidaili_config["password"]

        if settings.get("SMPPSIZE"):
            if settings.get("SMPPSIZE") < 1:
                raise ValueError("您输入的 ip 池子大小不能小于 1！请重新修改 settings 中的 SMPPSIZE 的值为大于等于 1")
            elif settings.get("SMPPSIZE") > 7:
                raise ValueError("你礼貌吗，ip 池最大数量不可以超过 7 个！请重新设置！")

            self.proxypool_size = settings.get("SMPPSIZE")
        else:
            self.proxypool_size = 1
        self.end_time = False
        self.time_count = np.zeros((self.proxypool_size,)) + 0.5
        self.reqnum_count = np.zeros((self.proxypool_size,))
        self.v_sum()
        self.t = Thread(target=self.clock, args=(spider,), daemon=True)
        self.t.start()
        self.get_proxy_ip(self.proxypool_size, True)
        spider.logger.info(f"初始化 ip 字典，ip 数量为：{len(self.proxy_list)}")
        spider.logger.info("SiMiProxyDownloaderMiddleware opened: %s" % spider.name)

    def spider_closed(self, spider):
        self.end_time = True
        self.t.join()
        spider.logger.info("SiMiProxyDownloaderMiddleware closed: %s" % spider.name)


class AbuDynamicProxyDownloaderMiddleware(object):
    """
    阿布云动态代理 - 隧道验证方式（其实和快代理的写法一致）
    """

    def __init__(self, settings):
        """
        从 scrapy 配置中取出动态隧道代理的信息
        Args:
            settings: scrapy 配置信息
        """
        dynamic_proxy_config = settings.get("DYNAMIC_PROXY_CONFIG", None)
        # 查看动态隧道代理配置是否符合要求
        is_match = ReuseOperation.if_dict_meet_min_limit(
            dict_config=dynamic_proxy_config,
            key_list=["PROXY_URL", "USERNAME", "PASSWORD"]
        )
        if not is_match:
            raise Exception(f"没有配置动态隧道代理，配置示例为：{Param.dynamic_proxy_config_example}")

        self.proxy_url = dynamic_proxy_config["PROXY_URL"]
        self.username = dynamic_proxy_config["USERNAME"]
        self.password = dynamic_proxy_config["PASSWORD"]

    @classmethod
    def from_crawler(cls, crawler):
        # This method is used by Scrapy to create your spiders.
        s = cls(crawler.settings)
        crawler.signals.connect(s.spider_opened, signal=signals.spider_opened)
        return s

    def spider_opened(self, spider):
        logger.info(f"阿布云动态隧道代理中间件: AbuDynamicProxyDownloaderMiddleware 已开启，生效脚本为: {spider.name}")

    def process_request(self, request, spider):
        if request.url.startswith("https://"):
            request.meta["proxy"] = "https://{}".format(self.proxy_url)
        elif request.url.startswith("http://"):
            request.meta["proxy"] = "http://{}".format(self.proxy_url)
        else:
            spider.logger.info(f"request url error: {request.url}")

        proxy_user_pass = self.username + ":" + self.password
        encoded_user_pass = "Basic " + base64.urlsafe_b64encode(bytes(proxy_user_pass, "ascii")).decode("utf8")
        request.headers["Proxy-Authorization"] = encoded_user_pass


class RequestByRequestsMiddleware(object):
    """
    将 scrapy 的 Request 请求替换为 requests
    """

    def process_request(self, request, spider):
        try:
            # 将 scrapy 的 request headers 的值，原封不动地转移到 requests 中
            r_headers = ToolsForAyu.get_dict_form_scrapy_req_headers(scrapy_headers=request.headers)

            request_body_str = str(request.body, encoding="utf-8")
            scrapy_request_method = str(request.method).upper()
            if scrapy_request_method == "GET":
                # 普通 GET 请求
                if not request.body:
                    r_response = requests.request(
                        method=scrapy_request_method,
                        url=request.url,
                        headers=r_headers,
                        cookies=request.cookies,
                        verify=False,
                        timeout=(Param.requests_req_timeout, Param.requests_res_timeout)
                    )

                # 携带 body 参数的 GET 请求
                else:
                    r_response = requests.request(
                        method=scrapy_request_method,
                        url=request.url,
                        headers=r_headers,
                        cookies=request.cookies,
                        data=request_body_str,
                        verify=False,
                        timeout=(Param.requests_req_timeout, Param.requests_res_timeout)
                    )

            elif scrapy_request_method == "POST":
                # POST 请求的情况是要判断其格式，如果是 json 的格式传来，则需要 requests 的 json.dumps 格式的 data 值
                if ReuseOperation.judge_str_is_json(judge_str=request_body_str):
                    r_response = requests.request(
                        method=scrapy_request_method,
                        url=request.url,
                        headers=r_headers,
                        data=request_body_str,
                        cookies=request.cookies,
                        verify=False,
                        timeout=(Param.requests_req_timeout, Param.requests_res_timeout)
                    )

                else:
                    post_data_dict = {x.split("=")[0]: x.split("=")[1] for x in request_body_str.split("&")}
                    r_response = requests.request(
                        method=scrapy_request_method,
                        url=request.url,
                        headers=r_headers,
                        data=post_data_dict,
                        cookies=request.cookies,
                        verify=False,
                        timeout=(Param.requests_req_timeout, Param.requests_res_timeout)
                    )
            else:
                raise Exception("出现未知请求方式，请及时查看！")
            return HtmlResponse(url=request.url, status=r_response.status_code, body=r_response.text, request=request, encoding="utf-8")

        except Exception as e:
            """这里比较重要，还是推荐捕获所有错误；不建议只抛错 requests.exceptions.ConnectTimeout:"""
            return HtmlResponse(url=request.url, status=202, body=f"requests 请求出现错误：{e}", request=request, encoding="utf-8")
